import os
import re
import sys
import time
from contextlib import contextmanager
from copy import copy, deepcopy
from operator import itemgetter
from os import path
from subprocess import Popen
from typing import Dict

from PyQt5.QtCore import QAbstractTableModel, QEvent, QItemSelection, QItemSelectionModel, QItemSelectionRange, \
    QMetaObject, QMimeData, QModelIndex, QPoint, QUrl, Qt, pyqtSignal, pyqtSlot, Q_ARG
from PyQt5.QtGui import QColor, QFont, QDrag, QPalette
from PyQt5.QtWidgets import QAbstractItemDelegate, QAbstractItemView, QAction, QApplication, QDialog, QGridLayout, QGroupBox, \
    QHBoxLayout, QHeaderView, QLabel, QLineEdit, QMenu, QMessageBox, QPushButton, QStyledItemDelegate, QTableView, \
    QVBoxLayout

from . import audioinfo
from .audioinfo import (PATH, FILENAME, DIRPATH, FILETAGS, READONLY, INFOTAGS, DIRNAME,
                        EXTENSION, FILENAME_NO_EXT, PARENT_DIR, CaselessDict, encode_fn, decode_fn, get_filename_tags)
from .puddleobjects import (unique, partial, natural_sort_key, gettag,
                            HeaderSetting, getfiles, progress, PuddleConfig, singleerror, winsettings, issubfolder,
                            fnmatch)

from collections import defaultdict
from .util import write, to_string
from .constants import SELECTIONCHANGED, SEPARATOR, BLANK
from . import confirmations
import logging
from .translations import translate
from .util import rename_error_msg
from .audio_filter import parse as filter_audio

the_break = False

status = {}

LIBRARY = '__library'
HIGHLIGHTCOLOR = Qt.GlobalColor.green
SHIFT_RETURN = 2
RETURN_ONLY = 1


def _default_audio_player():
    if sys.platform.startswith("linux"):
        return 'xdg-open'
    elif sys.platform == "darwin":
        return 'open -a iTunes'
    return "clementine -p"


def commontag(tag, tags):
    x = defaultdict(lambda: [])
    for audio in tags:
        if tag in audio:
            x[audio[tag][0]].append(audio)
        else:
            x[''].append(audio)
    return x


def not_in_dirs(files, dirs):
    if not dirs:
        return files
    for d in dirs:
        d = encode_fn(d)
        return [z for z in files if not z.startswith(d)]


def loadsettings(filepath=None):
    settings = PuddleConfig()
    if filepath:
        settings.filename = filepath
    titles = settings.get('tableheader', 'titles', [
        translate('Fields', 'Filename'), translate('Fields', 'Artist'),
        translate('Fields', 'Title'), translate('Fields', 'Album'),
        translate('Fields', 'Track'), translate('Fields', 'Length'),
        translate('Fields', 'Year'), translate('Fields', 'Bitrate'),
        translate('Fields', 'Genre'), translate('Fields', 'Comment'),
        translate('Fields', 'Dirpath')])
    tags = settings.get('tableheader', 'tags',
                        ['__filename', 'artist', 'title', 'album', 'track',
                         '__length', 'year', '__bitrate', 'genre', 'comment', '__dirpath'])
    # checked = settings.get('tableheader', 'enabled', range(len(tags)), True)
    # checked = []
    fontsize = settings.get('table', 'fontsize', 0, True)
    rowsize = settings.get('table', 'rowsize', -1, True)
    v1_option = settings.get('id3tags', 'v1_option', 2)
    v2_option = settings.get('id3tags', 'v2_option', 4)
    audioinfo.id3.v1_option = v1_option
    audioinfo.id3.v2_option = v2_option
    filespec = ';'.join(settings.get('table', 'filespec', []))

    return (list(zip(titles, tags)), fontsize, rowsize, filespec)


def caseless(tag, audio):
    if tag in audio:
        return tag
    smalltag = tag.lower()
    try:
        return [z for z in audio if isinstance(z, str) and z.lower() == smalltag][0]
    except IndexError:
        return tag


def has_previews(tags=None, parent=None, msg=None):
    """Disables preview mode. Returns True if disable successful, else False.

    Displays a confirmation to the user (message box) asking if they want
    to exit Preview Mode if any files contain previews. If the user selects
    that they want to exit, then True is returned. Preview Mode is not
    changed.

    If not previews in tags, then True is returned.

    tags is a list of audioinfo.Tag object with the preview attribute.
        If preview not empty, the file will be considered to have a preview.
        If not specified status['alltags'] is used.
    parent => A window to have as the parent for the message box.
    msg => Message to show in message box.
    """

    if msg is None:
        msg = translate("Previews", 'Do you want to exit Preview Mode? Unsaved changes will be discarded')

    if tags is None:
        tags = status['alltags']

    previews = False
    for z in tags:
        if z.preview:
            previews = True
            break

    if previews and confirmations.should_show('preview_mode'):
        ret = QMessageBox.question(parent, 'puddletag', msg)
        if ret != QMessageBox.StandardButton.Yes:
            return True
    return False


def tag_in_file(tag, audio):
    if tag in audio:
        return tag
    smalltag = tag.lower()
    try:
        return [z for z in audio if z.lower() == smalltag][0]
    except IndexError:
        return None


def model_tag(model, base=audioinfo.AbstractTag):
    class ModelTag(base):
        def __init__(self, *args, **kwargs):
            self.preview = CaselessDict()
            self.edited = None
            super(ModelTag, self).__init__(*args, **kwargs)

        @property
        def images(self):
            if not self.IMAGETAGS:
                return []
            if model.previewMode and '__image' in self.preview:
                return self.preview['__image']
            else:
                return base.images.fget(self)

        @images.setter
        def images(self, value):
            if not self.IMAGETAGS:
                return
            if model.previewMode:
                self.preview['__image'] = value
            else:
                base.images.fset(self, value)

        def __contains__(self, key):
            if model.previewMode:
                if key in self.preview:
                    return True
            return base.__contains__(self, key)

        def clear(self):
            self.preview.clear()
            super(ModelTag, self).clear()

        def clean(self):
            for k, v in self.preview.items():
                real = self.realvalue(k, '')
                if not isinstance(real, str):
                    real = ''.join(real)

                if not isinstance(v, str):
                    v = ''.join(v)

                if real == v:
                    del (self.preview[k])

        def __delitem__(self, key):
            if key in INFOTAGS:
                return
            if model.previewMode and key in self.preview:
                del (self.preview[key])
            else:
                super(ModelTag, self).__delitem__(key)

        def delete(self, tag=None):
            if model.previewMode:
                return
            try:
                base.delete(self, tag)
            except TypeError:
                base.delete(self)

        def equal_fields(self):
            if not (model.previewMode and self.preview):
                return []
            ret = []
            for k, v in self.preview.items():
                if k == '__image':
                    continue
                real = self.realvalue(k, '')
                if not isinstance(real, str):
                    real = ''.join(real)

                if not isinstance(v, str):
                    v = ''.join(v)

                if real == v:
                    ret.append(k)
            return ret

        def __getitem__(self, key):
            if model.previewMode and key in self.preview:
                return self.preview[key]
            else:
                return super(ModelTag, self).__getitem__(key)

        def keys(self):
            keys = list(super(ModelTag, self).keys())
            if model.previewMode and self.preview:
                [keys.append(key) for key in self.preview
                 if key not in keys]
            return keys

        def __len__(self):
            return len(list(self.keys()))

        def realvalue(self, key, default=None):
            try:
                return base.__getitem__(self, key)
            except KeyError:
                if default is not None:
                    return default
                raise

        def __setitem__(self, key, value):
            if model.previewMode:
                if key in FILETAGS:
                    value = to_string(value)

                if key not in READONLY:
                    if not value and key in self.preview:
                        del (self.preview[key])
                    else:
                        self.preview[key] = value
                    if key in FILETAGS:
                        self.update_file_fields_in_preview(key)
            else:
                super(ModelTag, self).__setitem__(key, value)

        def update_file_fields_in_preview(self, changed_key):
            """ Reimplements the setter-logic of the matching properties in MockTag, but for the preview."""
            if not changed_key or changed_key == PATH:
                filepath = self[PATH]
            elif changed_key in (DIRPATH, FILENAME):
                filepath = path.join(self[DIRPATH], self[FILENAME])
            elif changed_key in (FILENAME_NO_EXT, EXTENSION):
                filepath = path.join(self[DIRPATH], self[FILENAME_NO_EXT])
                if self[EXTENSION]:
                    filepath += os.path.extsep + self[EXTENSION]
            elif changed_key == DIRNAME:
                filepath = path.join(path.dirname(self[DIRPATH]), self[DIRNAME], self[FILENAME])
            else:
                return

            # clear all file fields from the preview
            for k in FILETAGS:
                if k in self.preview:
                    del self.preview[k]
            # and then write only those that actually are different
            for k, v in get_filename_tags(filepath).items():
                if self[k] != v:
                    self.preview[k] = v

        def update(self, dictionary=None):
            if not dictionary:
                return
            if model.previewMode:
                for k, v in dictionary.items():
                    self[k] = v
            else:
                super().update(dictionary)

    return ModelTag


def _Tag(model):
    splitext = path.splitext
    extensions = audioinfo.extensions

    options = [[Kind[0], model_tag(model, Kind[1]), Kind[2]] for Kind
               in audioinfo.options]
    filetypes = dict([(z[0], z) for z in options])

    extension_regex = re.compile(r'\.(%s)$' % '|'.join(extensions))

    def ReplacementTag(filename):

        try:
            fileobj = open(filename, "rb")
        except IOError:
            logging.info("Can't open file %s", filename)
            return None

        match = extension_regex.search(filename)
        if match:
            return filetypes[extensions[match.groups()[0]][0]][1](filename)

        try:
            header = fileobj.read(128)
            results = [Kind[0].score(filename, fileobj, header) for Kind in options]
        finally:
            fileobj.close()
        results = list(zip(results, options))
        results.sort(key=lambda x: x[0])
        score, Kind = results[-1]

        if score > 0:
            return Kind[1](filename)
        else:
            return None

    return ReplacementTag


class Properties(QDialog):
    def __init__(self, info, parent=None):
        """Shows a window with the properties in info.

        info should be a list of 2-entry tuples. These tuples should consists
        of a string in the 0-th index to be used as the title of a group.
        The first index (the properties) should also be a list length 2 tuples.
        Both indexes containing strings. Where the 0-th is used as the description
        and the other as the value.

        .e.g
        [('File', [('Filename', '/Hip Hop Songs/Nas-These Are Our Heroes .mp3'),
                  ('Size', '6151 kB'),
                  ('Path', 'Nas - These Are Our Heroes .mp3'),
                  ('Modified', '2009-07-28 14:04:05'),
                  ('ID3 Version', 'ID3v2.4')]),
        ('Version', [('Version', 'MPEG 1 Layer 3'),
                    ('Bitrate', '192 kb/s'),
                    ('Frequency', '44.1 kHz'),
                    ('Mode', 'Stereo'),
                    ('Length', '4:22')])]
        """
        QDialog.__init__(self, parent)
        winsettings('fileinfo', self)
        self.setWindowTitle(translate("Shortcut Settings", 'File Properties'))
        self._load(info)

    def _load(self, info):
        vbox = QVBoxLayout()
        interaction = Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.TextSelectableByKeyboard
        for title, items in info:
            frame = QGroupBox(title)
            framegrid = QGridLayout()
            framegrid.setColumnStretch(1, 1)
            for row, value in enumerate(items):
                prop = QLabel(value[0] + ':')
                prop.setTextInteractionFlags(interaction)
                framegrid.addWidget(prop, row, 0)
                propvalue = QLabel('<b>%s</b>' % value[1])

                propvalue.setTextInteractionFlags(interaction)
                framegrid.addWidget(propvalue, row, 1)
            frame.setLayout(framegrid)
            vbox.addWidget(frame)
        close = QPushButton('Close')
        close.clicked.connect(self.close)
        hbox = QHBoxLayout()
        hbox.addStretch()
        hbox.addWidget(close)
        vbox.addLayout(hbox)
        self.setLayout(vbox)


class ColumnSettings(HeaderSetting):
    """A dialog that allows you to edit the header of a TagTable widget."""
    title = translate("Column Settings", 'Columns')

    def __init__(self, parent=None, showok=False, status=None, table=None):

        self.tags, fontsize, rowsize, filespec = loadsettings()

        checked = []

        if table is None and status is not None:
            table = status['table']

        if table is not None:
            tags = table.model().headerdata
            hd = table.horizontalHeader()

            tags = ([hd.visualIndex(i), z, hd.isSectionHidden(i)] for i, z in
                    enumerate(tags))
            tags = sorted(tags)
            checked = [i for i, t in enumerate(tags) if not t[2]]
            tags = [tuple(z[1]) for z in tags]
            [tags.append(z) for z in self.tags if z not in tags]
            self.tags = tags

        HeaderSetting.__init__(self, self.tags, parent, showok, True)

        if showok:
            winsettings('columnsettings', self)

        self.setWindowFlags(Qt.WindowType.Widget)
        label = QLabel(translate("Column Settings", 'Adjust visibility of columns.'))
        self.grid.addWidget(label, 0, 0)
        items = [self.listbox.item(z) for z in range(self.listbox.count())]

        if not checked:
            checked = list(range(len(tags)))
        [z.setCheckState(Qt.CheckState.Checked) if i in checked
         else z.setCheckState(Qt.CheckState.Unchecked) for i, z in enumerate(items)]

    def applySettings(self, control=None):
        row = self.listbox.currentRow()
        if row > -1:
            self.tags[row][0] = str(self.textname.text())
            self.tags[row][1] = str(self.tag.currentText())
        checked = [z for z in range(self.listbox.count()) if
                   self.listbox.item(z).checkState()]
        titles = [z[0] for z in self.tags]
        tags = [z[1] for z in self.tags]
        cparser = PuddleConfig()
        cparser.set('tableheader', 'titles', titles)
        cparser.set('tableheader', 'tags', tags)

        hidden = [i for i, z in enumerate(self.tags) if i not in checked]

        if control:
            control.horizontalHeader().reset()
            control.setHeaderTags(self.tags, hidden)
            control.restoreSelection()
            control.horizontalHeader().reset()
        else:
            self.headerChanged.emit(self.tags, hidden)

    def add(self):
        HeaderSetting.add(self)
        self.listbox.currentItem().setCheckState(True)

    def duplicate(self):
        item = self.listbox.currentItem()
        if item:
            checked = item.checkState()
        HeaderSetting.duplicate(self)
        self.listbox.currentItem().setCheckState(checked)

    def okClicked(self):
        self.applySettings()
        self.close()


class TagModel(QAbstractTableModel):
    """The model used in TableShit
    Methods you shoud take not of are(read docstrings for more):

    setData -> As per the usual model, can only write one tag at a time.
    setRowData -> Writes a row's tags at once.
    undo -> undo's changes
    setTestData and unSetTestData -> Used to display temporary values in the table.
    """
    fileChanged = pyqtSignal(name='fileChanged')
    aboutToSort = pyqtSignal(name='aboutToSort')
    sorted = pyqtSignal(name='sorted')
    previewModeChanged = pyqtSignal(bool, name='previewModeChanged')
    dirsmoved = pyqtSignal(list, name='dirsmoved')
    libfilesedited = pyqtSignal(list, name='libfilesedited')
    setDataError = pyqtSignal(str, name='setDataError')
    enableUndo = pyqtSignal(bool, name='enableUndo')

    def __init__(self, headerdata, taginfo=None):
        """Load tags.

        headerdata must be a list of tuples
        where the first item is the displayrole and the second
        the tag to be used as in [("Artist", "artist"), ("Title", title")].

        taginfo is a list of audioinfo.Tag objects."""

        QAbstractTableModel.__init__(self)
        self.alternateAlbumColors = True
        self._headerData = []
        self.headerdata = headerdata
        self.colorRows = []
        self.sortOrder = (0, Qt.SortOrder.AscendingOrder)
        self.saveModification = True
        self._filtered = []
        self._previewMode = False
        self._prevhighlight = []
        self._permah = []
        self.permaColor = QColor(Qt.GlobalColor.green)
        self._colored = []
        audioinfo.Tag = _Tag(self)
        audioinfo.model_tag = partial(model_tag, self)
        self.showToolTip = True
        status['previewmode'] = False
        self._previewBackground = None
        self._selectionBackground = None
        self._undo = defaultdict(lambda: {})
        self._previewUndo = defaultdict(lambda: {})
        self.sortFields = []
        self.reverseSort = True
        self._savedundolevel = None

        with self.reset_context():
            if taginfo is not None:
                self.taginfo = unique(taginfo)
                self.sortByFields(self.sortFields, reverse=self.reverseSort)
            else:
                self.taginfo = []
            for z in self.taginfo:
                z.preview = {}
                z.previewundo = {}
                z.undo = {}
                z.color = None
                z._temp = {}
                if not hasattr(z, 'library'):
                    z.library = None
        self.undolevel = 0
        self._fontSize = QFont().pointSize()

    @property
    def fontSize(self):
        return self._fontSize

    @fontSize.setter
    def fontSize(self, size):
        self._fontSize = size
        top = self.index(self.rowCount(), 0)
        bottom = self.index(self.rowCount() - 1, self.columnCount() - 1)
        self.dataChanged.emit(top, bottom)

    @property
    def headerdata(self):
        return self._headerData

    @headerdata.setter
    def headerdata(self, tags):

        new_tags = []
        indexes = {}

        for i, t in enumerate(tags):
            t = tuple(t)
            if t not in indexes:
                indexes[t] = i
                new_tags.append(t)

        self._headerData = new_tags

    @property
    def columns(self) -> Dict[int, str]:
        """Return a mapping from column index to tag field of that column."""
        return {idx: field for idx, (title, field) in enumerate(self.headerdata)}

    @property
    def previewBackground(self):
        return self._previewBackground

    @previewBackground.setter
    def previewBackground(self, val):
        self._previewBackground = val
        rows = [i for i, z in enumerate(self.taginfo) if z.preview]
        if rows:
            top = self.index(min(rows), 0)
            bottom = self.index(max(rows), self.columnCount() - 1)
            self.dataChanged.emit(top, bottom)

    @property
    def previewMode(self):
        return self._previewMode

    @previewMode.setter
    def previewMode(self, value):
        if not value and self._previewMode:
            rows = []
            for row, audio in enumerate(self.taginfo):
                if audio.preview:
                    audio.preview = {}
                    rows.append(row)
            if rows:
                top = self.index(min(rows), 0)
                bottom = self.index(max(rows), self.columnCount() - 1)
                self.dataChanged.emit(top, bottom)

            self.undolevel = self._savedundolevel
            self._previewUndo.clear()
        else:
            self._savedundolevel = self.undolevel
        self._previewMode = value
        status['previewmode'] = value
        self.previewModeChanged.emit(value)

    @property
    def selectionBackground(self):
        return self._selectionBackground

    @selectionBackground.setter
    def selectionBackground(self, val):
        self._selectionBackground = val
        rows = [i for i, z in enumerate(self.taginfo) if z.color]
        if rows:
            top = self.index(min(rows), 0)
            bottom = self.index(max(rows), self.columnCount() - 1)
            self.dataChanged.emit(top, bottom)

    @property
    def undolevel(self):
        return self._undolevel

    @undolevel.setter
    def undolevel(self, value):
        if value == 0:
            self.enableUndo.emit(False)
            if self.previewMode:
                self._previewUndo.clear()
            else:
                self._undo.clear()
        else:
            self.enableUndo.emit(True)
        self._undolevel = value

    def _addUndo(self, audio, undo):
        if self.previewMode:
            self._previewUndo[self.undolevel][audio] = undo
        else:
            self._undo[self.undolevel][audio] = undo

    def applyFilter(self, pattern=None, matchcase=True):
        with self.reset_context():
            self.taginfo = self.taginfo + self._filtered
            taginfo = self.taginfo
            if (not pattern) and (not self._filtered):
                return
            elif not pattern:
                self._filtered = []
                return

            filtered = [(filter_audio(a, pattern), a) for a in self.taginfo]
            self._filtered = [z[1] for z in filtered if not z[0]]
            self.taginfo = [z[1] for z in filtered if z[0]]

    def changeFolder(self, olddir, newdir):
        """Used for changing the directory of all the files in olddir to newdir.
        i.e. All children of olddir will now become children of newdir

        No actual moving is done though."""

        olddir = encode_fn(olddir)

        folder = itemgetter(DIRPATH)
        tags = [(i, z) for i, z in enumerate(self.taginfo)
                if z.dirpath.startswith(olddir)]
        libtags = []
        for i, audio in tags:
            if audio.dirpath == olddir:
                audio.dirpath = newdir
            elif issubfolder(olddir, audio.dirpath):  # Newdir is a parent
                audio.dirpath = os.path.join(newdir, audio.dirname)
            if audio.library:
                audio.save(True)
        rows = [z[0] for z in tags]
        if rows:
            self.dataChanged.emit(
                self.index(min(rows), 0), self.index(max(rows),
                                                     self.columnCount() - 1))

    def columnCount(self, index=QModelIndex()):
        return len(self.headerdata)

    def _toString(self, val):
        if isinstance(val, str):
            return val.replace('\n', ' ')
        else:
            try:
                return '\\\\'.join(val).replace('\n', ' ')
            except TypeError:
                return str(val)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        row = index.row()
        if not index.isValid() or not (0 <= row < len(self.taginfo)):
            return None

        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.ToolTipRole, Qt.ItemDataRole.EditRole):
            try:
                audio = self.taginfo[row]
                tag = self.headerdata[index.column()][1]
                val = self._toString(audio[tag])
            except (KeyError, IndexError):
                return None

            if role == Qt.ItemDataRole.ToolTipRole:
                if not self.showToolTip:
                    return None
                if self.previewMode and \
                        audio.preview and tag in audio.preview:
                    try:
                        real = self._toString(audio.realvalue(tag))
                        if not real:
                            real = BLANK
                    except KeyError:
                        real = BLANK
                    if real != val:
                        tooltip = translate('Table', "Preview: {}\nReal: {}").format(val, self._toString(real))
                    else:
                        tooltip = val
                else:
                    tooltip = val
                return tooltip
            return val
        elif role == Qt.ItemDataRole.BackgroundRole:
            audio = self.taginfo[row]
            if audio.color:
                return audio.color
            elif self.previewMode and audio.preview:
                return self.previewBackground
        elif role == Qt.ItemDataRole.FontRole:

            field = self.headerdata[index.column()][1]
            f = QFont()
            if f.pointSize() != self.fontSize:
                f.setPointSize(self.fontSize)
            audio = self.taginfo[row]
            if field in audio.preview:
                real = audio.realvalue(field, '')
                if self._toString(audio[field]) != self._toString(real):
                    f.setBold(True)
                else:
                    f.setItalic(True)
            return f
        return None

    def deleteTag(self, row=None, audio=None, delete=True):
        if row is not None:
            audio = self.taginfo[row]
        tags = audio.usertags
        try:
            tags['__image'] = audio['__image']
        except KeyError:
            pass
        if delete:
            audio.delete()
        self._addUndo(audio, tags)

    def deleteTags(self, rows):
        [self.deleteTag(row) for row in rows]
        self.undolevel += 1

    def dropMimeData(self, data, action, row, column, parent=QModelIndex()):
        return True

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled
        return Qt.ItemFlag(QAbstractTableModel.flags(self, index) |
                            Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDropEnabled)

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.TextAlignmentRole:
            if orientation == Qt.Orientation.Horizontal:
                return int(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
            return int(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if orientation == Qt.Orientation.Horizontal:
            try:
                return str(self.headerdata[section][0])
            except IndexError:
                return None
        return int(section + 1)

    def highlight(self, rows):
        rows = rows[::]
        hcolor = self.selectionBackground
        nolight = set(self._prevhighlight).difference(rows)
        self._prevhighlight = rows[::]
        rows.extend(nolight)
        taginfo = self.taginfo

        def set_color(row):
            if row < len(taginfo):
                if row in nolight:
                    if taginfo[row] in self._colored:
                        taginfo[row].color = self.permaColor
                    else:
                        taginfo[row].color = None
                else:
                    taginfo[row].color = hcolor

        [set_color(row) for row in rows]
        if rows:
            top = self.index(min(rows), 0)
            bottom = self.index(max(rows), self.columnCount() - 1)
            self.dataChanged.emit(top, bottom)

    def insertColumns(self, column, count, parent=QModelIndex(), data=None):
        self.beginInsertColumns(parent, column, column + count)
        if data:
            self.headerdata.extend(data)
        else:
            self.headerdata += [("", "") for z in range(count - column + 1)]
        self.endInsertColumns()
        return True

    def reloadTags(self, tags):
        for z in tags:
            z.preview = {}
            z.undo = {}
            z.previewundo = {}
            z.color = None
            z._temp = {}
            if not hasattr(z, 'library'):
                z.library = None

        fns = dict((f.filepath, i) for i, f in enumerate(self.taginfo))

        to_append = []

        num_rows_to_insert = 0

        first = self.rowCount()
        for t in tags:
            if t.filepath in fns:
                self.taginfo[fns[t.filepath]] = t
            else:
                self.taginfo.append(t)
                num_rows_to_insert = 1

        self.beginInsertRows(QModelIndex(), first, len(self.taginfo) - 1)
        self.endInsertRows()

    def load(self, taginfo, headerdata=None, append=False):
        """Loads tags as in __init__.
        If append is true, then the tags are just appended."""
        if headerdata is not None:
            self.headerdata = headerdata

        for z in taginfo:
            z.preview = {}
            z.undo = {}
            z.previewundo = {}
            z.color = None
            z._temp = {}
            if not hasattr(z, 'library'):
                z.library = None

        if append:
            for field in self.sortFields:
                getter = lambda audio: audio.get(field, '')
                taginfo.sort(key=lambda a: natural_sort_key(a.get(field, '')), reverse=self.reverseSort)

            filenames = [z.filepath for z in self.taginfo]
            self.taginfo.extend([z for z in taginfo if z.filepath
                                 not in filenames])

            first = self.rowCount()
            self.beginInsertRows(QModelIndex(), first, first + len(taginfo) - 1)
            self.endInsertRows()

            top = self.index(first, 0)
            bottom = self.index(self.rowCount() - 1, self.columnCount() - 1)
            self.dataChanged.emit(top, bottom)
        else:
            with self.reset_context():
                self.taginfo = taginfo
                self._filtered = []
            self.sortByFields(self.sortFields, reverse=self.reverseSort)
            [setattr(z, 'color', None) for z in self.taginfo if z.color]
            self.undolevel = 0

    def moveRows(self, rows, row):

        taginfo = self.taginfo
        rows = sorted(rows)

        tags = [taginfo[i] for i in rows]

        num_moved = len(rows)

        first = min(rows)
        first = first if first < row else row

        last = max(rows)
        last = last if last > row else row

        top = self.index(first, 0)
        bottom = self.index(last, self.columnCount() - 1)

        while rows:
            del (taginfo[rows[0]])
            rows = [i - 1 for i in rows[1:]]

        if row >= len(taginfo):
            row = row - num_moved
            taginfo.extend(tags)

        [taginfo.insert(row, z) for z in reversed(tags)]

        self.dataChanged.emit(top, bottom)

    def permaHighlight(self, rows):
        rows = rows[::]
        hcolor = self.permaColor
        nolight = set(self._permah).difference(rows)
        self._permah = rows[::]
        rows.extend(nolight)
        taginfo = self.taginfo

        def set_color(row):
            if row < len(taginfo):
                if row in nolight:
                    taginfo[row].color = None
                else:
                    taginfo[row].color = hcolor

        [set_color(row) for row in rows]
        if rows:
            top = self.index(min(rows), 0)
            bottom = self.index(max(rows), self.columnCount() - 1)
            self.dataChanged.emit(top, bottom)

    def setColors(self, tags):
        for f in self._colored:
            f.color = None
        for tag in tags:
            tag.color = self.permaColor
        index = self.taginfo.index

        def row_index(tag):
            try:
                return index(tag)
            except ValueError:
                return len(self.taginfo)

        rows = list(map(row_index, tags + self._colored))
        if rows:
            top = self.index(min(rows), 0)
            bottom = self.index(max(rows), self.columnCount() - 1)
            self.dataChanged.emit(top, bottom)
        self._colored = tags

    def removeColumns(self, column, count, parent=QModelIndex()):
        self.beginRemoveColumns(QModelIndex(), column, column + count - 1)
        del (self.headerdata[column: column + count])
        self.endRemoveColumns()
        return True

    def removeFolders(self, folders, v=True):
        if v:
            f = [i for i, tag in enumerate(self.taginfo) if tag.dirpath
                 not in folders and not tag.library]
        else:
            f = [i for i, tag in enumerate(self.taginfo) if tag.dirpath
                 in folders and not tag.library]
        while f:
            try:
                self.removeRows(f[0])
                del (f[0])
                f = [z - 1 for z in f]
            except IndexError:
                break

    @pyqtSlot(int, result=bool)
    def removeRows(self, position, rows=1, index=QModelIndex()):
        """Please, only use this function to remove one row at a time. For some reason, it doesn't work
        too well on debian if more than one row is removed at a time."""
        self.beginRemoveRows(QModelIndex(), position,
                             position + rows - 1)
        del (self.taginfo[position])
        self.endRemoveRows()
        return True

    @contextmanager
    def reset_context(self):
        self.beginResetModel()
        try:
            yield None
        finally:
            self.endResetModel()

    def rowColors(self, rows=None, clear=False):
        """Changes the background of rows to green.

        If rows is None, then the background of all the rows in the table
        are returned to normal."""
        taginfo = self.taginfo
        if rows:
            if clear:
                for row in rows:
                    if hasattr(taginfo[row], "color"):
                        del (taginfo[row].color)
            else:
                for row in rows:
                    taginfo[row].color = HIGHLIGHTCOLOR
                for i, z in enumerate(taginfo):
                    if i not in rows and hasattr(taginfo[row], "color"):
                        del (taginfo[row].color)
        else:
            rows = []
            for i, tag in enumerate(self.taginfo):
                if hasattr(tag, 'color'):
                    del (tag.color)
                    rows.append(i)
        if rows:
            firstindex = self.index(min(rows), 0)
            lastindex = self.index(max(rows), self.columnCount() - 1)
            self.dataChanged.emit(firstindex, lastindex)

    def rowCount(self, index=QModelIndex()):
        return len(self.taginfo)

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole, dontwrite=False):
        """Sets the data of the currently edited cell as expected.
        Also writes tags and increases the undolevel."""
        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        if index.isValid() and 0 <= index.row() < len(self.taginfo):
            column = index.column()
            tag = self.headerdata[column][1]
            currentfile = self.taginfo[index.row()]
            newvalue = str(value)
            realtag = currentfile.mapping.get(tag, tag)
            if realtag in FILETAGS and tag not in [FILENAME, EXTENSION, DIRNAME]:
                QApplication.restoreOverrideCursor()
                return False

            if tag not in FILETAGS:
                newvalue = [_f for _f in newvalue.split(SEPARATOR) if _f]
            if newvalue == currentfile.get(tag, '') and not dontwrite:
                QApplication.restoreOverrideCursor()
                return False
            if dontwrite:
                QApplication.restoreOverrideCursor()
                return {tag: newvalue}, index.row()
            ret = self.setRowData(index.row(), {tag: newvalue}, undo=True)
            if not self.previewMode and currentfile.library:
                self.libfilesedited.emit([ret])
            self.undolevel += 1
            self.fileChanged.emit()
            QApplication.restoreOverrideCursor()
            return True
        return False

    def setHeaderData(self, section, orientation, value, role=Qt.ItemDataRole.EditRole):
        if (orientation == Qt.Orientation.Horizontal) and (role == Qt.ItemDataRole.DisplayRole):
            self.headerdata[section] = value

        self.headerdata = self.headerdata  # make sure columns are set

        self.headerDataChanged.emit(orientation, section, section)

    def setHeader(self, tags):
        with self.reset_context():
            self.headerdata = tags
            self.headerDataChanged.emit(
                Qt.Orientation.Horizontal, 0, len(self.headerdata))

    def setRowData(self, row, tags, undo=False, justrename=False, temp=False):
        """A function to update one row.
        row is the row, tags is a dictionary of tags.

        If undo`is True, then an undo level is created for this file.
        If justrename is True, then (if tags contain a PATH or EXTENSION key)
        the file is just renamed i.e not tags are written.
        """
        audio = self.taginfo[row]

        temporary = temp
        ret = None
        oldpath = audio.dirpath

        if self.previewMode:
            if temporary:
                for field in tags:
                    if field not in audio._temp:
                        audio._temp[field] = audio.get(field, '')
            preview = audio.preview
            undo_val = dict([(tag, copy(audio[tag])) if tag in audio
                             else (tag, []) for tag in tags])
            audio.update(tags)
            if undo:
                if audio._temp:
                    undo_val.update(audio._temp)
                    audio._temp = {}
                self._addUndo(audio, undo_val)
            return
        else:
            artist = audio.get('artist', '')
            try:
                undo_val = write(audio, tags, self.saveModification, justrename)
            except PermissionError as e:
                raise e
            if undo and undo_val:
                self._addUndo(audio, undo_val)

            if DIRNAME in tags or DIRPATH in tags:
                self.changeFolder(oldpath, audio.dirpath)
                self.dirsmoved.emit([[oldpath, audio.dirpath]])
            if justrename and audio.library:
                audio.save(True)
            ret = (artist, tags)

        return ret

    def setTestData(self, rows, previews=None):
        """A method that allows you to change the visible data of
        the model without writing tags.

        rows is the rows that you want to change
        previews -> is the tags that are to be shown."""
        taginfo = self.taginfo
        if rows and (not previews):
            index = taginfo.index
            tags = rows
            previews = []
            rows = []
            for audio, preview in tags.items():
                rows.append(index(audio))
                previews.append(preview)
        if not rows:
            return
        if not self.previewMode:
            return
        unsetrows = rows[len(previews):]

        if unsetrows:
            self.unSetTestData(rows=unsetrows)
        for row, preview in zip(rows, previews):
            taginfo[row].preview.clear()
            taginfo[row].update(preview)
        firstindex = self.index(min(rows), 0)
        lastindex = self.index(max(rows), self.columnCount() - 1)
        self.dataChanged.emit(firstindex, lastindex)
        self.fileChanged.emit()

    def sibling(self, row, column, index=QModelIndex()):
        if row < (self.rowCount()) and row >= 0:
            return self.index(row, column)
        return QModelIndex()

    def sort(self, column, order=Qt.SortOrder.DescendingOrder):
        try:
            field = self.headerdata[column][1]
        except IndexError:
            if len(self.headerdata) >= 1:
                field = self.headerdata[0][1]
            else:
                return
        if order == Qt.SortOrder.DescendingOrder:
            self.sortByFields([field], reverse=True)
        else:
            self.sortByFields([field], reverse=False)

    def sortByFields(self, fields, files=None, rows=None, reverse=None):

        self.aboutToSort.emit()

        if reverse is None and fields == self.sortFields:
            reverse = not self.reverseSort
        elif reverse is None:
            reverse = False

        with self.reset_context():
            if files and rows:
                for field in fields:
                    files.sort(key=lambda a: natural_sort_key(a.get(field, '')), reverse=reverse)
                for index, row in enumerate(rows):
                    self.taginfo[row] = files[index]
            else:
                for field in fields:
                    self.taginfo.sort(key=lambda a: natural_sort_key(a.get(field, '')), reverse=reverse)

            self.reverseSort = reverse
            self.sortFields = fields

        self.sorted.emit()

    def supportedDropActions(self):
        return Qt.DropAction.CopyAction

    def undo(self):
        """Undos the last action.

        Basically, if a tag has a key which is = self.undolevel - 1,
        then the tag is updated with the dictionary in that key.

        setRowData does not modify the undoleve unless you explicitely tell
        it, but setData does modify the undolevel.

        It is recommended that you use consecutive indexes for self.undolevel."""
        if self.undolevel <= 0:
            self.undolevel = 0
            return
        level = self.undolevel - 1
        oldfiles = []
        newfiles = []
        rows = []
        edited = []
        if self.previewMode:
            undo = self._previewUndo
        else:
            undo = self._undo
        if level not in undo:
            return

        get_row = self.taginfo.index

        for audio, undo_tags in undo[level].items():
            row = get_row(audio)
            rows.append(row)
            if self.previewMode:
                audio.update(undo_tags)
                for k, v in undo_tags.items():
                    if v == [] and k in audio.preview:
                        del (audio.preview[k])
            else:
                if audio.library:
                    oldfiles.append(deepcopy(audio.tags))
                edited.append(self.setRowData(row, undo_tags, False))

        del (undo[level])
        if rows:
            self.updateTable(rows)
            if edited:
                self.libfilesedited.emit(edited)
        if self.undolevel > 0:
            self.undolevel -= 1

    def unSetTestData(self, rows=None):
        taginfo = self.taginfo
        if not rows:
            rows = [i for i, z in enumerate(taginfo) if z.preview]
            if not rows:
                return
        for row in rows:
            taginfo[row].preview = {}

        if rows:
            firstindex = self.index(min(rows), 0)
            lastindex = self.index(max(rows), self.columnCount() - 1)
            self.dataChanged.emit(firstindex, lastindex)

    def updateTable(self, rows):
        firstindex = self.index(min(rows), 0)
        lastindex = self.index(max(rows), self.columnCount() - 1)
        self.dataChanged.emit(firstindex, lastindex)


class TagDelegate(QStyledItemDelegate):
    def createEditor(self, parent, option, index):
        # logging.debug('Creating editor')
        editor = QLineEdit(parent)
        editor.returnPressed = False
        editor.writeError = False
        editor.setFrame(False)
        font = editor.font()
        font.setPointSize(index.model().fontSize)
        editor.setFont(font)

        # logging.debug('Closing editor.')
        return editor

    def eventFilter(self, editor, event):
        if event.type() == QEvent.Type.KeyPress:
            if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
                if event.key() == Qt.Key.Key_Return:
                    shift_pressed = event.modifiers() == Qt.KeyboardModifier.ShiftModifier
                else:
                    shift_pressed = event.modifiers() == Qt.KeyboardModifier.ShiftModifier | Qt.KeyboardModifier.KeypadModifier
                if shift_pressed:
                    editor.returnPressed = SHIFT_RETURN
                else:
                    editor.returnPressed = RETURN_ONLY
        return QStyledItemDelegate.eventFilter(self, editor, event)

    def setModelData(self, editor, model, index):
        try:
            model.setData(index, editor.text())
        except EnvironmentError as e:
            editor.writeError = e


class TableHeader(QHeaderView):
    """A headerview put here simply to enable the contextMenuEvent
    so that I can show the edit columns menu.

    Call it with tags in the usual form, to set the top header."""
    saveSelection = pyqtSignal(name='saveSelection')
    headerChanged = pyqtSignal([list, list], name='headerChanged')

    def __init__(self, orientation, parent=None):
        QHeaderView.__init__(self, orientation, parent)
        self.setVisible(True)
        self.setSectionsClickable(True)
        self.setHighlightSections(True)
        self.setSectionsMovable(True)
        self.setSortIndicatorShown(True)
        self.setSortIndicator(0, Qt.SortOrder.AscendingOrder)

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        settings = menu.addAction(
            translate("Column Settings", "&Select Columns"))
        settings.triggered.connect(self.setTitles)

        menu.exec_(event.globalPos())

    def mousePressEvent(self, event):
        if event.button == Qt.MouseButton.RightButton:
            self.contextMenuEvent(event)
            return
        QHeaderView.mousePressEvent(self, event)

    def setTitles(self):
        self.win = ColumnSettings(showok=True, table=self.parent())
        self.win.setModal(True)
        self.win.show()
        self.win.headerChanged.connect(self.headerChanged)


class VerticalHeader(QHeaderView):
    def __init__(self, orientation, parent=None):
        QHeaderView.__init__(self, orientation, parent)
        self.setDefaultSectionSize(self.minimumSectionSize() + 3)
        self.setMinimumSectionSize(1)

        self.sectionResized.connect(self._resize)

    def _resize(self, row, oldsize, newsize):
        self.setDefaultSectionSize(newsize)


class TagTable(QTableView):
    """I need a more descriptive name for this.

    This table is the table that handles all my tags for me.
    The main functions and properties are:

    rowTags(row) -> Returns the tags from a row.
    updateRow(row, tags) - > Updates a row with the tags specified
    selectedRows -> A list of currently selected rows
    remRows() -> Removes the selected rows.
    playcommand -> Command to run to play files.
    """
    dirschanged = pyqtSignal(list, name='dirschanged')
    tagselectionchanged = pyqtSignal(name='tagselectionchanged')
    filesloaded = pyqtSignal(bool, name='filesloaded')
    viewfilled = pyqtSignal(bool, name='viewfilled')
    filesselected = pyqtSignal(bool, name='filesselected')
    enableUndo = pyqtSignal(bool, name='enableUndo')
    playlistchanged = pyqtSignal(str, name='playlistchanged')
    deletedfromlib = pyqtSignal(list, name='deletedfromlib')
    libfilesedited = pyqtSignal(list, name='libfilesedited')
    previewModeChanged = pyqtSignal(bool, name='previewModeChanged')
    onetomany = pyqtSignal(dict, name='onetomany')
    setDataError = pyqtSignal(str, name='setDataError')
    dirsmoved = pyqtSignal(list, name='dirsmoved')
    itemSelectionChanged = pyqtSignal(name='itemSelectionChanged')

    def __init__(self, headerdata=None, parent=None):
        QTableView.__init__(self, parent)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.settingsdialog = ColumnSettings
        if not headerdata:
            headerdata = []
        header = TableHeader(Qt.Orientation.Horizontal, self)
        self.setSortingEnabled(True)
        self._savedSelection = False

        self.setVerticalHeader(VerticalHeader(Qt.Orientation.Vertical))
        self.setHorizontalHeader(header)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setAlternatingRowColors(True)
        self.setWordWrap(False)
        self.showmsg = True
        self._currentcol = {}
        self._currentrow = {}
        self._currenttags = []
        self.dirs = []
        self._resize = False
        self._restore = False
        self._selectedRows = []
        self._selectedColumns = []
        self._select = True
        self._playlist = False
        self.filespec = ''
        self.contextMenu = None
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        model = TagModel(headerdata)
        self.setModel(model)

        def emitundo(val):
            self.enableUndo.emit(val)
            self.selectionChanged()

        model.enableUndo.connect(emitundo)
        model.libfilesedited.connect(self.libfilesedited)
        self.undo = model.undo

        self.closeEditor = self._closeEditor
        delegate = TagDelegate(self)
        self.setItemDelegate(delegate)

        self.subFolders = False

        def sep():
            separator = QAction(self)
            separator.setSeparator(True)
            return separator

        self.emits = ['dirschanged', SELECTIONCHANGED, 'filesloaded',
                      'viewfilled', 'filesselected', 'enableUndo',
                      'playlistchanged', 'deletedfromlib', 'libfilesedited',
                      'previewModeChanged', 'onetomany']
        self.receives = [
            ('loadFiles', self.loadFiles),
            ('removeFolders', self.removeFolders),
            ('filter', self.applyFilter),
            ('setpreview', self.setTestData),
            ('loadtags', self.load_tags),
            ('highlight', self.highlight),
            ('enable_preview_mode', partial(self.previewMode, True)),
            ('disable_preview_mode', partial(self.previewMode, False)),
        ]

        self.gensettings = [
            ('Su&bfolders', True),
            ('Show &gridlines', True),
            ('Show tooltips in file-view:', True),
            ('Show &row numbers', True),
            ('Automatically resize columns to contents', False),
            ('&Preserve file modification times (if supported)', False),
            ('Program to &play files with:', _default_audio_player())
        ]

        status['selectedrows'] = lambda: self.selectedRows
        status['selectedfiles'] = lambda: self.selectedTags
        status['firstselection'] = self._firstSelection
        status['selectedcolumns'] = lambda: self.selectedColumns
        status['selectedtags'] = self._getSelectedTags
        status['alltags'] = lambda: self.model().taginfo
        status['allfiles'] = lambda: self.model().taginfo
        status['table'] = self

    @property
    def fontSize(self):
        return self.model().fontSize

    @fontSize.setter
    def fontSize(self, size):
        self.model().fontSize = size

    def increaseFont(self):
        self.fontSize += 1

    def decreaseFont(self):
        self.fontSize -= 1

    @property
    def autoresize(self):
        return self._resize

    @autoresize.setter
    def autoresize(self, value):
        self._resize = value
        if value:
            self.resizeColumnsToContents()
            self.model().modelReset.connect(self.resizeColumnsToContents)
        else:
            try:
                self.model().modelReset.disconnect(self.resizeColumnsToContents)
            except TypeError:
                logging.debug("Tried to disconnect un-connected Resize-Slot")

    def _getSelectedTags(self):
        columns = self.model().columns
        rows = self.currentRowSelection()
        audios = self.selectedTags

        def get_selected(f, row):
            selected = (columns[column] for column in rows[row])

            return ((field, f.get(field, '')) for field in selected)

        return list(map(dict, (get_selected(*z) for z in zip(audios, sorted(rows)))))

    def _firstSelection(self):
        if not self.selectedRows:
            raise IndexError

        get_index = self.model().index
        isselected = self.selectionModel().isSelected

        field_mapping = self.model().columns
        row = min(self.selectedRows)
        fields = [field_mapping[c] for c in range(self.columnCount())
                  if isselected(get_index(row, c))]

        audio = self.model().taginfo[row]

        return audio, dict((field, audio.get(field, '')) for field in fields)

    def applyGenSettings(self, d, startlevel=None):
        self.saveSelection()
        if not startlevel:
            self.subFolders = d['Su&bfolders']
            self.setShowGrid(d['Show &gridlines'])
            self.verticalHeader().setVisible(d['Show &row numbers'])
            self.autoresize = d['Automatically resize columns to contents']
            self.model().saveModification = d[
                '&Preserve file modification times (if supported)']
            self.playcommand = d['Program to &play files with:'].split(' ')
            self.model().showToolTip = d['Show tooltips in file-view:']

    def applyFilter(self, pattern=None, matchcase=True):
        self.saveSelection()
        self.clearSelection()
        self.model().applyFilter(pattern, matchcase)
        self.restoreSelection()

    def changeFolder(self, olddir, newdir, updatemodel=True):
        try:
            for i, d in enumerate(self.dirs[::]):
                if d == olddir:
                    self.dirs[i] = newdir
                elif d.startswith(olddir):
                    self.dirs[i] = newdir + d[len(olddir):]
        except IndexError:
            pass
        if updatemodel:
            self.model().changeFolder(olddir, newdir)

    def clearAll(self):
        with self.model().reset_context():
            self.model().taginfo = []
        self.dirschanged.emit([])

    def _closeEditor(self, editor, hint=QAbstractItemDelegate.EndEditHint.NoHint):
        if editor.writeError:
            model = self.model()
            currentfile = model.taginfo[self.currentIndex().row()]
            QTableView.closeEditor(self, editor, QAbstractItemDelegate.EndEditHint.NoHint)
            model.setDataError.emit(
                rename_error_msg(editor.writeError, currentfile.filepath))
            return

        if len(self.selectedRows) > 1:
            QTableView.closeEditor(self, editor, QAbstractItemDelegate.EndEditHint.NoHint)
        elif not editor.returnPressed:
            QTableView.closeEditor(self, editor, hint)
        else:
            index = self.currentIndex()
            newindex = None

            if editor.returnPressed == RETURN_ONLY:
                if index.row() < self.rowCount() - 1:
                    newindex = self.model().index(index.row() + 1,
                                                  index.column())
            elif editor.returnPressed == SHIFT_RETURN:
                if index.row() > 0:
                    newindex = self.model().index(index.row() - 1,
                                                  index.column())

            if newindex:
                QTableView.closeEditor(self, editor, QAbstractItemDelegate.EndEditHint.NoHint)
                self.setCurrentIndex(newindex)
                self.edit(newindex)
            else:
                QTableView.closeEditor(self, editor, QAbstractItemDelegate.EndEditHint.NoHint)

    def removeTags(self):
        if self.model().previewMode:
            QMessageBox.information(self, 'puddletag',
                                    translate("Table", 'Disable Preview Mode first to enable tag deletion.'))
            return
        deltag = self.model().deleteTag

        def func():
            for row in self.selectedRows:
                try:
                    deltag(row)
                    yield None
                except (OSError, IOError) as e:
                    msg = translate('Table',
                                    "An error occurred while deleting the tag of {}: <b>{}</b>"
                                    ).format(e.filename, e.strerror)
                    yield msg, len(self.selectedRows)
                except NotImplementedError as e:
                    f = self.model().taginfo[row]
                    filename = f[PATH]
                    ext = f[EXTENSION]
                    rowlen = len(self.selectedRows)
                    yield translate("Table", "There was an error deleting the "
                                             "tag of {}: <b>Tag deletion isn't supported "
                                             "for {} files.</b>").format(filename, ext), rowlen

        def fin():
            self.selectionChanged()
            self.model().undolevel += 1

        f = progress(func, translate("Table", 'Deleting tag... '),
                     len(self.selectedRows), fin)
        f(self)

    def columnCount(self):
        return self.model().columnCount()

    def rowCount(self):
        return self.model().rowCount()

    def contextMenuEvent(self, event):
        if self.contextMenu:
            self.contextMenu.exec_(event.globalPos())

    @property
    def isempty(self):
        if self.model().rowCount() <= 0:
            return True
        return False

    def deleteSelectedWithoutMessage(self):
        self.deleteSelected(showmsg=False)

    def deleteSelected(self, delfiles=True, ifgone=False, showmsg=None):
        if showmsg is None:
            showmsg = True
            showmsg = confirmations.should_show('delete_files')

        if delfiles and showmsg:
            result = QMessageBox.question(self, "puddletag",
                                          translate("Table", "Are you sure you want to delete the selected files?"))
        else:
            result = QMessageBox.StandardButton.Yes
        if result != QMessageBox.StandardButton.Yes:
            return
        selected = self.selectedTags
        selectedRows = self.selectedRows
        curindex = self.currentIndex()
        last = max(selectedRows) - len(selectedRows) + 1, curindex.column()
        libtags = []

        def func():
            temprows = sorted(selectedRows[::])
            for ((i, row), audio) in zip(enumerate(selectedRows), selected):
                try:
                    filename = audio.filepath
                    os.remove(filename)
                    if audio.library:
                        audio.remove()
                        libtags.append(audio)
                    # Cross-thread call
                    QMetaObject.invokeMethod(self.model(), 'removeRows',
                                             Q_ARG(int, temprows[i]),
                                             )
                    temprows = [z - 1 for z in temprows]
                    yield None
                except (OSError, IOError) as detail:
                    msg = "I couldn't delete <b>%s</b> (%s)" % (filename,
                                                                detail.strerror)
                    if row == temprows[-1]:
                        yield msg, 1
                    else:
                        yield msg, len(selectedRows)

        def fin():
            index = self.model().index(*last)
            if libtags:
                self.deletedfromlib.emit(libtags)
            if index.isValid():
                self.setCurrentIndex(index)
            else:
                self.selectionChanged()

        s = progress(func, translate("Table", 'Deleting '), len(selectedRows), fin)
        if self.parentWidget():
            s(self.parentWidget())
        else:
            s(self)

    def dragEnterEvent(self, event):
        event.accept()
        event.acceptProposedAction()

    def dropEvent(self, event):
        mime = event.mimeData()
        if event.source() == self and \
                hasattr(mime, 'draggedRows') and mime.draggedRows:

            row = self.rowAt(event.pos().y())
            if row == -1:
                row = self.rowCount() - 1
            self.saveSelection()
            self.model().moveRows(mime.draggedRows, row)
            self.restoreSelection()
        else:
            filenames = [z.path() for z
                         in event.mimeData().urls()]

            dirs = []
            files = []

            for f in filenames:
                if not f:
                    continue
                if os.path.isdir(f):
                    dirs.append(f)
                else:
                    files.append(f)

            if not dirs and not files:
                return

            self.loadFiles(files, dirs, append=True)

    def dragMoveEvent(self, event):
        mime = event.mimeData()
        if event.source() == self:
            if hasattr(mime, 'draggedRows') and mime.draggedRows:
                event.accept()
            else:
                event.ignore()
            return

        if mime.hasUrls():
            event.accept()
        else:
            event.ignore()

    def mouseMoveEvent(self, event):

        if event.buttons() != Qt.MouseButton.LeftButton:
            return
        mimeData = QMimeData()
        plainText = ""
        tags = []
        if hasattr(self, "selectedRows"):
            selectedRows = self.selectedRows[::]
        else:
            return
        try:
            pnt = QPoint(*self.StartPosition)
        except AttributeError:
            return
        if (event.pos() - pnt).manhattanLength() < QApplication.startDragDistance():
            return
        filenames = [z.filepath for z in self.selectedTags]
        urls = list(map(QUrl.fromLocalFile, list(map(decode_fn, filenames))))
        mimeData = QMimeData()
        mimeData.setUrls(urls)
        if event.modifiers() == Qt.KeyboardModifier.MetaModifier:
            mimeData.draggedRows = self.selectedRows[::]
        else:
            mimeData.draggedRows = None

        drag = QDrag(self)
        drag.setMimeData(mimeData)
        drag.setHotSpot(event.pos() - self.rect().topLeft())
        dropaction = drag.exec_()
        if dropaction == Qt.DropAction.MoveAction:
            if not os.path.exists(filenames[0]):
                self.deleteSelected(False, False, False)

    def mousePressEvent(self, event):
        if event.buttons() == Qt.MouseButton.LeftButton:
            self.StartPosition = [event.pos().x(), event.pos().y()]
        QTableView.mousePressEvent(self, event)

    def fillTable(self, tags, append=False):
        """Clears the table and fills it with metadata in tags.

        tags are a list of audioinfo.Tag objects.
        If append is True then the tags are just appended.
        """
        self.selectedRows = []
        self.selectedColumns = []
        if append is None:
            self.model().reloadTags(tags)
        else:
            self.model().load(tags, append=append)

        QApplication.processEvents()
        if append or append is None:
            self.restoreSelection()
        else:
            self.selectCorner()

        if self.autoresize:
            self.resizeColumnsToContents()
        if self.model().taginfo:
            self.viewfilled.emit(True)
        else:
            self.viewfilled.emit(False)

    def highlight(self, rows):
        self.model().setColors(rows)

    def invertSelection(self):
        model = self.model()
        topLeft = model.index(0, 0)
        bottomRight = model.index(model.rowCount() - 1, model.columnCount() - 1)

        selection = QItemSelection(topLeft, bottomRight)
        self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.Toggle)

    def keyPressEvent(self, event):
        event.accept()
        # You might think that this is redundant since a delete
        # action is defined in contextMenuEvent, but if this isn't
        # done then the delegate is entered.

        has_modifier = event.modifiers() in [Qt.KeyboardModifier.ControlModifier, Qt.KeyboardModifier.ShiftModifier, Qt.KeyboardModifier.ControlModifier | Qt.KeyboardModifier.ShiftModifier]
        if event.key() == Qt.Key.Key_Delete and self.selectedRows:
            self.deleteSelected()
            return
        # This is so that an item isn't edited when the user's holding the shift or
        # control key.
        elif event.key() == Qt.Key.Key_Space and (has_modifier):
            trigger = self.editTriggers()
            self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
            ret = QTableView.keyPressEvent(self, event)
            self.setEditTriggers(trigger)
            return ret
        return QTableView.keyPressEvent(self, event)

    def loadFiles(self, files=None, dirs=None, append=False, subfolders=None,
                  filepath=None, post_process=None):
        assert files or dirs, 'Either files or dirs (or both) must be specified.'

        if subfolders is None:
            subfolders = self.subFolders

        if not files:
            files = []
        if not dirs:
            dirs = []
        elif isinstance(dirs, str):
            dirs = [dirs]

        dirs = list(map(encode_fn, dirs))

        files = not_in_dirs(not_in_dirs(files, dirs), self.dirs)

        if subfolders:
            new_dirs = []
            for i, d in enumerate(dirs):
                if not [z for z in dirs if issubfolder(z, d, None)]:
                    new_dirs.append(d)

            dirs = new_dirs

        if append:
            self.saveSelection()
            if subfolders:
                # Remove all subfolders if the parent's already loaded.
                for d in self.dirs:
                    dirs = [z for z in dirs if not z.startswith(d)]
                toremove = set()
                for d in dirs:
                    toremove = toremove.union([z for z in self.dirs if z.startswith(d)])
                self.removeFolders(toremove)
            else:
                self.removeFolders([z for z in dirs if z in self.dirs], False)
            self.dirs.extend(dirs)
        else:
            self.dirs = dirs

        if not files and not dirs:
            return

        tags = []
        if len(dirs) == 1:
            reading_dir = translate('Defaults', "Reading Directory: {}").format(dirs[0])
        elif dirs:
            reading_dir = translate('Defaults', "Reading Directory: {} + others").format(dirs[0])
        else:
            reading_dir = translate('Defaults', "Reading Dir")

        def load_dir():
            if files:
                filenames = files
            else:
                filenames = []
            for fname in getfiles(dirs, subfolders):
                filenames.append(fname)
                yield reading_dir

            if self.filespec and self.filespec.strip():
                filenames = fnmatch(self.filespec, filenames)

            yield len(filenames)

            for fname in filenames:
                tag = gettag(fname)
                if tag is not None:
                    tags.append(tag)
                yield None

        if post_process:
            s = progress(load_dir, translate("Defaults", 'Loading '), 20,
                         lambda: post_process(tags, append, filepath))
        else:
            s = progress(load_dir, translate("Defaults", 'Loading '), 20,
                         lambda: self._loadFilesDone(tags, append, filepath))
        s(self.parentWidget())

    def _loadFilesDone(self, tags, append, filepath):
        self.fillTable(tags, append)
        if not filepath:
            self.dirschanged.emit(self.dirs[::])
        else:
            self.playlistchanged.emit(filepath)
            self.dirschanged.emit([])
        self.filesloaded.emit(True)
        if self._restore:
            self.restoreSelection(self._restore)

        model = self.model()
        start_index = model.index(0, 0)
        end_index = model.index(model.rowCount() - 1, model.columnCount() - 1)
        model.dataChanged.emit(start_index, end_index)
        self.selectionChanged()

    def load_tags(self, tags):
        self.fillTable(tags, False)
        self.dirs = []
        self.dirschanged.emit(self.dirs[::])
        self.filesloaded.emit(True)
        sortcolumn = self.horizontalHeader().sortIndicatorSection()
        QTableView.sortByColumn(self, sortcolumn, Qt.SortOrder.AscendingOrder)

    def loadSettings(self):
        (tags, fontsize, rowsize, self.filespec) = loadsettings()
        self.setHeaderTags(tags)
        if fontsize:
            self.fontSize = fontsize
        if rowsize > -1:
            self.verticalHeader().setDefaultSectionSize(rowsize)

        cparser = PuddleConfig()
        preview_color = cparser.get('table', 'preview_color', [192, 255, 192], True)
        default = QPalette().color(QPalette.ColorRole.Mid).getRgb()[:-1]
        selection_color = cparser.get('table', 'selected_color', default, True)

        model = self.model()

        model.previewBackground = QColor.fromRgb(*preview_color)
        model.selectionBackground = QColor.fromRgb(*selection_color)

        sort_fields = cparser.get('table', 'sort_fields', [])
        reverse = cparser.get('table', 'sort_reverse', False, True)

        # print sort_fields, model.sortFields

        if sort_fields:
            model.sortByFields(sort_fields, reverse=reverse)
        else:
            h = self.horizontalHeader()
            model.sort(h.sortIndicatorSection(), h.sortIndicatorOrder())

    def moveDown(self, rows=None):
        if rows is None:
            rows = self.selectedRows[::]
        if not rows:
            return
        elif max(rows) >= self.rowCount() - 1:
            return

        taginfo = self.model().taginfo

        self.saveSelection()
        self._select = False

        for row in reversed(rows):
            if row >= self.rowCount() - 1:
                continue
            old = taginfo[row]
            new = taginfo[row + 1]
            taginfo[row + 1] = old
            taginfo[row] = new

        getindex = self.model().index
        self.model().dataChanged.emit(getindex(min(rows), 0), getindex(max(rows), self.columnCount()))

        self.restoreSelection()
        self._select = True

    def moveUp(self, rows=None):
        if rows is None:
            rows = self.selectedRows[::]

        if not rows:
            return
        elif min(rows) <= 0:
            return

        taginfo = self.model().taginfo

        self.saveSelection()
        self._select = False

        for i, row in enumerate(rows):
            if row <= 0:
                continue
            old = taginfo[row]
            new = taginfo[row - 1]
            taginfo[row] = new
            taginfo[row - 1] = old

        getindex = self.model().index
        self.model().dataChanged.emit(getindex(min(rows), 0), getindex(max(rows), self.columnCount()))

        self.restoreSelection()
        self._select = True

    def playFiles(self):
        """Play the selected files using the player specified in self.playcommand"""
        if not self.selectedRows: return
        if hasattr(self, "playcommand"):

            li = list(map(encode_fn, self.playcommand))

            li.extend([z.filepath for z in self.selectedTags])

            try:
                Popen(li)
            except (OSError) as detail:
                if detail.errno != 2:
                    QMessageBox.critical(self, translate('Defaults', "Error"),
                                         translate('Table',
                                                   "An error occurred while trying to play the selected files: "
                                                   "<b>{}</b><br />Does the music player you defined (<b>{}</b>) "
                                                   "exist?").format(detail.strerror, " ".join(self.playcommand)))
                else:
                    QMessageBox.critical(self, translate('Defaults', "Error"),
                                         translate('Table',
                                                   "It wasn't possible to play the selected files, because the music "
                                                   "player you defined (<b>{}</b>) does not exist."
                                                   ).format(" ".join(self.playcommand)))

    def previewMode(self, value):
        if not value:
            if has_previews(self.model().taginfo, self):
                return False
        self.model().previewMode = value
        return value

    def reloadSelected(self, files=None):
        self._restore = self.saveSelection()

        loaded_dirs = list(map(encode_fn, self.dirs))
        if files is None:
            files = self.model().taginfo

        taginfo = self.model().taginfo
        dirs = set([taginfo[i].dirpath for i in self.selectedRows])

        is_sub = lambda fn: [_f for _f in [issubfolder(z, fn) for z in dirs] if _f]

        sub_files = []
        for f in files:
            if f.dirpath in dirs or is_sub(f.dirpath):
                sub_files.append(f)

        self.loadFiles(None, dirs, False, self.subFolders,
                       self._playlist,
                       lambda t, a, f: self.__processReload(t, a, f, sub_files))

    def __processReload(self, tags, append, filepath, sub_files):
        new_fns = set(z.filepath for z in tags)
        to_remove = set(z for z in sub_files
                        if z.filepath not in new_fns)
        with self.model().reset_context():
            self.model().taginfo = [z for z in self.model().taginfo if
                                    z not in to_remove]
            self.fillTable(tags, None)

        self.filesloaded.emit(True)
        if self._restore:
            self.restoreSelection(self._restore)

        self.selectionChanged()

    def reloadFiles(self, files=None):
        previews = False
        for z in self.model().taginfo:
            if z.preview:
                previews = True
                break
        if previews:
            ret = QMessageBox.question(self, 'puddletag',
                translate("Previews", 'There are unsaved changes pending. Do you want to discard and reload?'))
            if ret != QMessageBox.StandardButton.Yes:
                return

        self._restore = self.saveSelection()
        dirs = list(map(encode_fn, self.dirs))
        files = [z.filepath for z in self.model().taginfo if z.dirpath
                 not in dirs]
        libfiles = [z for z in self.model().taginfo if '__library' in z]
        if self._playlist:
            self.loadFiles(files, dirs, False, self.subFolders)
        else:
            self.loadFiles(files, dirs, False, self.subFolders,
                           self._playlist)
        self.model().load(libfiles, append=True)

    def rowTags(self, row, stringtags=False):
        """Returns all the tags pertinent to the file at row."""
        if stringtags:
            return self.model().taginfo[row].stringtags()
        return self.model().taginfo[row]

    def selectCurrentColumn(self):
        if self.selectedIndexes():
            col = self.currentIndex().column()
            model = self.model()
            topLeft = model.index(0, col)
            bottomRight = model.index(model.rowCount() - 1, col)

            selection = QItemSelection(topLeft, bottomRight)
            self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.Select)

    def selectCorner(self):
        topLeft = self.model().index(0, 0)
        selection = QItemSelection(topLeft, topLeft)
        self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.Select)
        self.setCurrentIndex(topLeft)

    def setModel(self, model):
        QTableView.setModel(self, model)
        self.updateRow = model.setRowData
        model.modelReset.connect(self.selectionChanged)
        model.setDataError.connect(self.writeError)
        model.fileChanged.connect(self.selectionChanged)
        model.aboutToSort.connect(self.saveBeforeReset)
        model.sorted.connect(self.restoreSort)
        model.previewModeChanged.connect(self.previewModeChanged)
        model.dirsmoved.connect(self.dirsmoved)
        set_data = model.setData

        def modded_setData(index, value, role=Qt.ItemDataRole.EditRole):
            if len(self.selectedRows) == 1:
                return set_data(index, value, role)
            ret = set_data(index, value, role, True)

            if ret:
                self.onetomany.emit(ret[0])
            return False

        model.setData = modded_setData

    def currentRowSelection(self):
        """Returns a dictionary with the currently selected rows as keys.
        Each key contains a list with the selected columns of that row.

        {} is returned if nothing is selected."""
        x = {}
        for z in self.selectedIndexes():
            try:
                x[z.row()].append(z.column())
            except KeyError:
                x[z.row()] = [z.column()]
        return x

    def currentColumnSelection(self):
        x = {}
        for z in self.selectedIndexes():
            try:
                x[z.column()].append(z.row())
            except KeyError:
                x[z.column()] = [z.row()]
        return x

    @property
    def selectedTags(self):
        rowTags = self.rowTags
        return [rowTags(row) for row in self.selectedRows]

    @property
    def selectedRows(self):
        return self._selectedRows

    @selectedRows.setter
    def selectedRows(self, val):
        self._selectedRows = val

    @property
    def selectedColumns(self):
        return self._selectedColumns

    @selectedColumns.setter
    def selectedColumns(self, val):
        self._selectedColumns = val

    def selectionChanged(self, selected=None, deselected=None):
        """Pretty important. This updates self.selectedRows, which is used
        everywhere.

        I've set selected an deselected as None, because I sometimes
        want self.selectedRows updated without hassle."""
        t = time.time()
        if selected is not None and deselected is not None:
            QTableView.selectionChanged(self, selected, deselected)
        taginfo = self.model().taginfo
        model = self.model()

        selectedRows = set()
        selectedColumns = set()

        for z in self.selectedIndexes():
            selectedRows.add(z.row())
            selectedColumns.add(z.column())
        self.selectedRows = sorted(list(selectedRows))
        self.selectedColumns = sorted(list(selectedColumns))

        if self._select:
            self.itemSelectionChanged.emit()
            if self.selectedRows:
                self.filesselected.emit(True)
            else:
                self.filesselected.emit(False)
            model.highlight(self.selectedRows)
            self.tagselectionchanged.emit()

    def saveBeforeReset(self):
        self.setCursor(Qt.CursorShape.BusyCursor)
        self._savedSelection = self.saveSelection()

    def restoreSort(self):
        self.setCursor(Qt.CursorShape.ArrowCursor)
        if self._savedSelection:
            self.restoreSelection(self._savedSelection)
            self._savedSelection = None
            return

    def saveSelection(self):
        taginfo = self.model().taginfo
        fields = [field for title, field in self.model().headerdata]
        filenames = defaultdict(lambda: set())
        selection = self.currentRowSelection()
        for row, columns in self.currentRowSelection().items():
            filenames[taginfo[row].filepath] = set(fields[c] for c in columns)
        self.__savedSelection = filenames
        return filenames

    def restoreReloadSelection(self, currentrow, currentcol, tags):
        if not tags:
            return

        def getGroups(rows):
            groups = []
            try:
                last = [rows[0]]
            except IndexError:
                return []
            for row in rows[1:]:
                if row - 1 == last[-1]:
                    last.append(row)
                else:
                    groups.append(last)
                    last = [row]
            groups.append(last)
            return groups

        modelindex = self.model().index
        filenames = [z.filepath for z in self.model().taginfo]
        getrow = lambda x: filenames.index(x.filepath)
        selection = QItemSelection()
        select = lambda top, low, col: selection.append(
            QItemSelectionRange(modelindex(top, col),
                                modelindex(low, col)))

        newindexes = {}
        while True:
            try:
                tag = tags[0]
            except IndexError:
                break
            try:
                newindexes[tag[0]] = getrow(tag[1])
            except ValueError:
                pass
            del (tags[0])

        groups = {}
        for col, rows in currentcol.items():
            groups[col] = getGroups(sorted([newindexes[row] for row in rows if row in newindexes]))

        for col, rows in groups.items():
            [select(min(row), max(row), col) for row in rows]
        self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.Select)

    def restoreSelection(self, data=None):
        if data is None:
            data = self.__savedSelection

        get_index = self.model().index
        selection = QItemSelection()

        def select_index(row, col):
            model_index = get_index(row, col)
            selection.select(model_index, model_index)

        columns = dict((field[1], i) for i, field in
                       enumerate(self.model().headerdata))

        for row, fn in enumerate(z.filepath for z in self.model().taginfo):
            selected_fields = data.get(fn)
            if not selected_fields:
                continue

            for field in selected_fields:
                column = columns.get(field)
                if column is not None:
                    select_index(row, column)

        self.selectionModel().clear()
        self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.SelectCurrent)
        self.model().highlight(self.selectedRows)

    def removeFolders(self, dirs, valid=True):
        if dirs:
            self.dirs = list(set(self.dirs).difference(dirs))
            self.model().removeFolders(dirs, valid)

    def setHeaderTags(self, tags, hidden=None):

        self.saveSelection()
        old_column_sizes = {val[0]: self.horizontalHeader().sectionSize(idx)
            for idx,val in enumerate(self.model().headerdata)
            if self.horizontalHeader().sectionSize(idx) > 0}

        self.model().setHeader(tags)

        hd = TableHeader(Qt.Orientation.Horizontal, self)
        self.setHorizontalHeader(hd)
        if not self.autoresize:
            for idx, val in enumerate(self.model().headerdata):
                if val[0] in old_column_sizes:
                    hd.resizeSection(idx, old_column_sizes[val[0]])
        if hidden is not None:
            for c in hidden:
                hd.hideSection(c)

        self.restoreSelection()

    def setHorizontalHeader(self, header):
        header.setModel(self.model())
        QTableView.setHorizontalHeader(self, header)
        header.headerChanged.connect(self.setHeaderTags)
        header.saveSelection.connect(self.saveSelection)

    def writeError(self, text):
        """Shows a tooltip when an error occors.

        Actually, a tooltip is never shown, because the table
        is updated as soon as it tries to show it. So a setDataError
        signal is emitted with the text that can be used to show
        text in the status bar or something."""
        singleerror(self.parentWidget(), text)
        self.setDataError.emit(text)

    def saveSettings(self):
        cparser = PuddleConfig()
        cparser.set('table', 'fontsize', self.fontSize)
        rowsize = self.verticalHeader().defaultSectionSize()
        cparser.set('table', 'rowsize', rowsize)
        cparser.set('table', 'sort_fields', self.model().sortFields)
        cparser.set('table', 'sort_reverse', self.model().reverseSort)

    def nextDir(self, dirpaths, previous=False):
        taginfo = self.model().taginfo
        row = self.currentIndex().row()
        column = self.currentIndex().column()
        i = 0
        d = taginfo[row].dirpath
        while True:
            try:
                if d != taginfo[row + i].dirpath:
                    new_dir = taginfo[row + i].dirpath
                    break
            except IndexError:
                return
            if not previous:
                i += 1
                if row + i >= len(taginfo):
                    row = 0 - i
            else:
                i -= 1

        return dict((row, column) for row in dirpaths[new_dir])

    def selectPrevDir(self):
        return self.selectDir(True)

    def selectRow(self, row):
        selection = QItemSelection()
        get_index = self.model().index
        columns = self.selectedColumns
        for column in columns:
            index = get_index(row, column)
            selection.merge(QItemSelection(index, index),
                            QItemSelectionModel.SelectionFlag.Select)
        self.selectionModel().select(selection,
                                     QItemSelectionModel.SelectionFlag.ClearAndSelect)

        self.scrollTo(get_index(row, min(columns)), QAbstractItemView.ScrollHint.EnsureVisible)

    def selectDir(self, previous=False):
        model = self.model()
        modelindex = model.index
        selection = QItemSelection()
        merge = selection.merge
        taginfo = model.taginfo

        selected = defaultdict(lambda: [])
        selected_rows = set()

        for index in self.selectedIndexes():
            if index.row() not in selected_rows:
                selected[taginfo[index.row()].dirpath].append(index.column())
                selected_rows.add(index.row())

        dirpaths = defaultdict(lambda: set())

        dirs_sorted = []

        for row, f in enumerate(taginfo):
            dirpaths[f.dirpath].add(row)
            dirs_sorted.append(f.dirpath)

        to_select = {}

        for d, columns in selected.items():
            if len(columns) == len(dirpaths[d]) and len(selected) == 1:
                to_select = self.nextDir(dirpaths, previous)
                if to_select:
                    self.selectionModel().clearSelection()
                    self.setCurrentIndex(
                        modelindex(min(to_select), min(to_select.values())))
                else:
                    return
            else:
                for row in dirpaths[d]:
                    to_select[row] = columns[-1]

        for row, column in to_select.items():
            index = modelindex(row, column)
            merge(QItemSelection(index, index), QItemSelectionModel.SelectionFlag.Select)

        self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.Select)

    def showProperties(self):
        f = self.selectedTags[0]
        win = Properties(f.info, self.parentWidget())
        win.show()

    def setPlayCommand(self, command):
        self.playcommand = command

    def setTestData(self, data):
        if hasattr(data, 'items'):
            self.model().setTestData(data)
        else:
            self.model().setTestData(self.selectedRows, data)

    def sortByColumn(self, column, order):
        """Guess"""
        QTableView.sortByColumn(self, column, order)
        self.restoreSelection()

    def wheelEvent(self, e):
        h = self.horizontalScrollBar()
        if not self.verticalScrollBar().isVisible() and h.isVisible():
            numsteps = e.angleDelta().y() // 5
            h.setValue(h.value() - numsteps)
            e.accept()
        else:
            QTableView.wheelEvent(self, e)



import json
import os
import pickle
from collections import OrderedDict

from PyQt5.QtCore import QByteArray, QMimeData, pyqtSignal, QObject
from PyQt5.QtWidgets import QApplication

from .. import actiondlg
from .. import findfunc
from .. import helperwin
from ..puddleobjects import (PuddleConfig, PuddleDock)

from ..audioinfo import PATH, DIRPATH, FILETAGS, tag_to_json, encode_fn, decode_fn
from .. import musiclib, about as about
from ..util import split_by_tag, translate, to_string
from .. import functions
from .tagtools import *
from ..constants import HOMEDIR, SEPARATOR

path = os.path

status = {}

_TAGS_MIME_TYPE = 'application/x.puddletag.tags'


def applyaction(files=None, funcs=None):
    if files is None:
        files = status['selectedfiles']
    try:
        if isinstance(funcs[0], findfunc.Macro):
            r = findfunc.apply_macros
        else:
            r = findfunc.apply_actions
    except IndexError:
        return
    state = {'__total_files': str(len(files))}
    state['__files'] = files

    def func():
        for i, f in enumerate(files):
            yield r(funcs, f, state)

    emit('writeaction', func(), None, state)


def applyquickaction(files, funcs):
    if isinstance(funcs[0], findfunc.Macro):
        qa = findfunc.apply_macros
    else:
        qa = findfunc.apply_actions

    selected = status['selectedtags']
    state = {'__total_files': str(len(selected))}
    t = (qa(funcs, f, state, list(s.keys())) for f, s in zip(files, selected))
    emit('writeselected', t)


def auto_numbering(parent=None):
    """Shows the autonumbering wizard and sets the tracks
        numbers should be filled in"""
    tags = status['selectedfiles']
    numtracks = len(tags)

    win = helperwin.AutonumberDialog(parent, 1, numtracks, False)
    win.setModal(True)
    win.newtracks.connect(partial(number_tracks, tags, parent))
    win.show()


def clipboard_to_tag(parent=None):
    win = helperwin.ImportTextFile(parent, clipboard=True)
    win.setModal(True)
    win.patterncombo.addItems(status['patterns'])

    cparser = PuddleConfig()
    last_dir = cparser.get('importwindow', 'lastdir', HOMEDIR)
    win.lastDir = last_dir
    last_pattern = cparser.get('importwindow', 'lastpattern', '')
    if last_pattern:
        win.patterncombo.setEditText(last_pattern)

    def fin_edit(taglist, pattern):
        cparser.set('importwindow', 'lastdir', win.lastDir)
        cparser.set('importwindow', 'lastpattern', pattern)
        emit('writeselected', taglist)

    win.Newtags.connect(fin_edit)

    win.show()


def connect_status(actions):
    connect = lambda a: getattr(obj, a.status).connect(a.setStatusTip)
    actions = [x for x in actions if x.status]
    list(map(connect, actions))


def _setClipboardTags(tags):
    """ Save the given list of tags in the clipboard, both as json and binary. """
    mime = QMimeData()

    binaryData = pickle.dumps(tags)
    mime.setData(_TAGS_MIME_TYPE, binaryData)

    # This is only to support pasting outside of puddletag
    jsonData = json.dumps(list(map(tag_to_json, tags)))
    mime.setText(jsonData)

    QApplication.clipboard().setMimeData(mime)


def copy():
    """ Save the selected tags to the clipboard. """
    selected = status['selectedtags']
    _setClipboardTags(selected)


def copy_whole():
    """ Save all tags of the selected files to the clipboard. """
    tags = []
    mime = QMimeData()

    def usertags(f, images=True):
        ret = f.usertags
        if images and hasattr(f, 'images') and f.images:
            ret.update({'__image': f.images})
        return ret

    tags = [usertags(f) for f in status['selectedfiles']]

    data = json.dumps(list(map(tag_to_json, tags)))

    to_copy = check_copy_data(data)
    if to_copy == 2:
        tags = [usertags(f, False) for f in status['selectedfiles']]
    elif to_copy == 1:
        return

    _setClipboardTags(tags)


def cut():
    """ Copy the selected tags in the clipboard, and set them blank. """
    selected = status['selectedtags']
    _setClipboardTags(selected)

    emit('writeselected', (dict([(z, "") for z in s if z not in FILETAGS])
                           for s in selected))


def check_copy_data(data):
    """ If the given JSON data is bigger than 5MiB, show a dialog.
        Returns 0 if the user wants to copy all, 1 if not, and 2 if
        all except the images."""
    if len(data) > 5 * 1024 * 1024:
        msgbox = QMessageBox()
        msgbox.setText(translate("Messages",
                                 "That's a large amount of data to copy.\n"
                                 "It may cause your system to lock up.\n\n"
                                 "Do you want to go ahead?"))

        msgbox.setIcon(QMessageBox.Icon.Question)
        yesBtn = msgbox.addButton(translate("Defaults", "&Yes"),
                                  QMessageBox.ButtonRole.YesRole)
        noBtn = msgbox.addButton(translate("Defaults", "No"),
                                 QMessageBox.ButtonRole.NoRole)
        msgbox.setDefaultButton(noBtn)
        msgbox.setEscapeButton(noBtn)
        noImgBtn = msgbox.addButton(translate("Messages", "Copy without images."),
                                    QMessageBox.ButtonRole.ApplyRole)

        msgbox.exec_()

        if msgbox.clickedButton() is yesBtn:
            return 0
        if msgbox.clickedButton() is noBtn:
            return 1
        elif msgbox.clickedButton() is noImgBtn:
            return 2
    else:
        return 0


def display_tag(tag):
    """Used to display tags in the status bar in a human parseable format."""
    if not tag:
        return "<b>Error: Pattern does not match filename.</b>"
    s = "%s: <b>%s</b>, "
    tostr = lambda i: i if isinstance(i, str) else i[0]
    return "".join([s % (z, tostr(v)) for z, v in tag.items()])[:-2]


def extended_tags(parent=None):
    rows = status['selectedrows']
    win = helperwin.ExTags(files=status['selectedfiles'], parent=parent)
    win.setModal(True)
    win.rowChanged.connect(status['table'].selectRow)
    win.loadSettings()
    x = lambda val: emit('onetomany', val)
    win.extendedtags.connect(x)
    win.show()


def filename_to_tag():
    tags = status['selectedfiles']
    pattern = status['patterntext']

    x = [findfunc.filenametotag(pattern, tag[PATH], True)
         for tag in tags]
    emit('writeselected', x)


def format(parent=None, preview=None):
    """Formats the selected tags."""
    files = status['selectedfiles']
    pattern = status['patterntext']
    selected = status['selectedtags']

    ret = []
    tf = findfunc.tagtofilename

    state = {'__total_files': str(len(files))}
    for i, (audio, s) in enumerate(zip(files, selected)):
        state['__counter'] = str(i + 1)
        val = tf(pattern, audio, state=state)
        ret.append(dict([(tag, val) for tag in s]))
    emit('writeselected', ret)


def in_lib(state, parent=None):
    if state:
        if not status['library']:
            QMessageBox.critical(parent, translate("MusicLib", 'No libraries found'),
                                 translate("MusicLib", "Load a lib first."))
            return False
        files = status['allfiles']
        lib = status['library']
        libartists = status['library'].artists
        to_highlight = []
        for artist, tracks in split_by_tag(files, 'artist', None).items():
            if artist in libartists:
                libtracks = lib.get_tracks('artist', artist)
                titles = [track['title'][0].lower()
                          if 'title' in track else '' for track in libtracks]
                for track in tracks:
                    if track.get('title', [''])[0].lower() in titles:
                        to_highlight.append(track)
        emit('highlight', to_highlight)
        return True
    else:
        emit('highlight', [])
        return False


def load_musiclib(parent=None):
    try:
        m = musiclib.LibChooseDialog(parent)
    except musiclib.MusicLibError:
        QMessageBox.critical(parent, translate("MusicLib", 'No libraries found'),
                             translate("MusicLib", "No supported music libraries were found. Most likely "
                                                   "the required dependencies aren't installed. Visit the "
                                                   "puddletag website, <a href='http://puddletag.sourceforge.net'>"
                                                   "puddletag.sourceforge.net</a> for more details."))
        return
    m.setModal(True)
    m.adddock.connect(emit_received('adddock'))
    m.show()


def _pad(trknum, total, padlen):
    if total is not None:
        text = str(trknum).zfill(padlen) + "/" + str(total).zfill(padlen)
    else:
        text = str(trknum).zfill(padlen)
    return text


def number_tracks(tags, parent, offset, numtracks, restartdirs, padlength, split_field='__dirpath', output_field='track', by_group=False):
    """Numbers the selected tracks sequentially in the range
    between the indexes.
    The first item of indices is the starting track.
    The second item of indices is the number of tracks."""

    if not split_field:
        QMessageBox.critical(parent, translate("Autonumbering Wizard", 'Field empty...'),
                             translate("Autonumbering Wizard",
                                       "The field specified to use as a directory splitter was invalid. "
                                       "Please check your values."))
        return

    if not output_field:
        QMessageBox.critical(parent, translate("Autonumbering Wizard", 'Field empty...'),
                             translate("Autonumbering Wizard",
                                       "The output field specified was invalid. "
                                       "Please check your values."))
        return

    if restartdirs:  # Restart dir numbering
        folders = OrderedDict()
        for tag_index, tag in enumerate(tags):
            key = findfunc.parsefunc(split_field, tag)
            if not isinstance(key, str):
                key = tag.stringtags().get(split_field)

            if key in folders:
                folders[key].append(tag_index)
            else:
                folders[key] = [tag_index]
    else:
        folders = {'fol': [i for i, t in enumerate(tags)]}

    taglist = {}
    for group_num, tags in enumerate(folders.values()):
        if numtracks == -2:
            total = len(tags)
        elif numtracks == -1:
            total = None
        elif numtracks >= 0:
            total = numtracks
        for trknum, index in enumerate(tags):
            if by_group:
                trknum = group_num + offset
            else:
                trknum += offset

            text = _pad(trknum, total, padlength)
            taglist[index] = {output_field: text}

    taglist = [v for k, v in sorted(list(taglist.items()), key=lambda x: x[0])]

    emit('writeselected', taglist)


def _getClipboardTags():
    """ Returns the list of tags from the clipboard (if any) """
    if not QApplication.clipboard().mimeData().hasFormat(_TAGS_MIME_TYPE):
        return None
    binaryData = QApplication.clipboard().mimeData().data(_TAGS_MIME_TYPE).data()
    tags = pickle.loads(binaryData)
    return tags


def paste():
    """ Write tags from the clipboard to the selected files. """
    rows = status['selectedrows']
    if not rows:
        return
    clip = _getClipboardTags()
    if not clip:
        return
    tags = []
    while len(tags) < len(rows):
        tags.extend(clip)
    tags.extend(clip)
    emit('writeselected', tags)


def paste_onto():
    """ Write the tag values from the clipboard into the selected tags. """
    clip = _getClipboardTags()
    if not clip:
        return
    selected = status['selectedtags']
    tags = []
    while len(tags) < len(selected):
        tags.extend(clip)
    emit('writeselected', (dict(list(zip(s, list(cliptag.values()))))
                           for s, cliptag in zip(selected, tags)))


def rename_dirs(parent=None):
    """Changes the directory of the currently selected files, to
    one as per the pattern in self.patterncombo."""
    selectedfiles = status['selectedfiles']
    audio, selected = status['firstselection']
    pattern = status['patterntext']
    if not pattern:
        return

    func = findfunc.Function('tag_dir')
    func.args = [pattern]
    func.tag = ['sthaoeusnthaoeusnthaosnethu']

    run_func(selectedfiles, func)


def run_action(parent=None, quickaction=False):
    files = status['selectedfiles']
    if files:
        example = files[0]
    else:
        example = {}

    if quickaction:
        tags = status['selectedtags'][0]
        win = actiondlg.ActionWindow(parent, example, list(tags.keys()))
    else:
        win = actiondlg.ActionWindow(parent, example)
    win.setModal(True)

    if quickaction:
        func = partial(applyquickaction, files)
        win.donewithmyshit.connect(func)
    else:
        func = partial(applyaction, files)
        win.donewithmyshit.connect(func)
    action_tool = PuddleDock._controls['Actions']
    win.actionOrderChanged.connect(action_tool.updateOrder)
    if not quickaction:
        win.checkedChanged.connect(action_tool.updateChecked)
    win.show()


def run_function(parent=None, prevfunc=None):
    selectedfiles = status['selectedfiles']

    if not prevfunc:
        prevfunc = status['prevfunc']

    example = selectedfiles[0]
    try:
        selected_file = status['selectedtags'][0]
        selected_fields = list(selected_file.keys())
        text = selected_file[selected_fields[0]]
    except IndexError:
        text = example.get('title')

    if prevfunc:
        f = actiondlg.CreateFunction(prevfunc=prevfunc, parent=parent,
                                     selected_fields=selected_fields, example=example, text=text)
    else:
        f = actiondlg.CreateFunction(parent=parent,
                                     selected_fields=selected_fields, example=example, text=text)

    f.valschanged.connect(partial(run_func, selectedfiles))
    f.setModal(True)
    f.show()


def run_func(selectedfiles, func):
    status['prevfunc'] = func
    selectedtags = status['selectedtags']

    function = func.runFunction
    state = {'__total_files': str(len(selectedtags))}

    def tagiter():
        for i, (selected, f) in enumerate(zip(selectedtags, selectedfiles)):
            state['__counter'] = str(i + 1)
            fields = findfunc.parse_field_list(func.tag, f, list(selected.keys()))
            rowtags = f.tags
            ret = {}
            for field in fields:
                val = function(rowtags.get(field, ''), rowtags, state, r_tags=f)
                if val is not None:
                    if isinstance(val, str) and SEPARATOR in val:
                        val = val.split(SEPARATOR)
                    if hasattr(val, 'items'):
                        ret.update(val)
                    else:
                        ret[field] = val
            yield ret

    emit('writeselected', tagiter())


run_quick_action = lambda parent=None: run_action(parent, True)


def search_replace(parent=None):
    selectedfiles = status['selectedfiles']
    audio, selected = status['firstselection']

    try:
        text = to_string(list(selected.values())[0])
    except IndexError:
        text = translate('Defaults', '')

    func = findfunc.Function('replace')
    func.args = [text, text, False, False]
    func.tag = ['__selected']

    dialog = actiondlg.CreateFunction(prevfunc=func, parent=parent,
                                      selected_fields=list(selected.keys()), example=audio, text=text)

    dialog.valschanged.connect(partial(run_func, selectedfiles))
    dialog.setModal(True)
    dialog.controls[0].combo.setFocus()
    dialog.show()


def show_about(parent=None):
    win = about.AboutPuddletag(parent)
    win.setModal(True)
    win.exec_()


def tag_to_file():
    pattern = status['patterntext']
    files = status['selectedfiles']

    tf = functions.move
    state = {'__total_files': str(len(files))}

    def rename():
        for i, f in enumerate(files):
            state['__counter'] = str(i + 1)
            yield tf(f, pattern, f, state=state)

    emit('writeselected', rename())


def text_file_to_tag(parent=None):
    dirpath = status['selectedfiles'][0].dirpath

    win = helperwin.ImportTextFile(parent)
    cparser = PuddleConfig()
    last_dir = cparser.get('importwindow', 'lastdir', HOMEDIR)
    if win.openFile(dirpath=last_dir):
        win.close()
        return
    win.setModal(True)
    win.patterncombo.addItems(status['patterns'])

    last_pattern = cparser.get('importwindow', 'lastpattern', '')
    if last_pattern:
        win.patterncombo.setEditText(last_pattern)

    def fin_edit(taglist, pattern):
        cparser.set('importwindow', 'lastdir', win.lastDir)
        cparser.set('importwindow', 'lastpattern', pattern)
        emit('writeselected', taglist)

    win.Newtags.connect(fin_edit)
    win.show()


def update_status(enable=True):
    files = status['selectedfiles']
    pattern = status['patterntext']
    tf = lambda *args, **kwargs: encode_fn(findfunc.tagtofilename(*args, **kwargs))
    if not files:
        return
    tag = files[0]

    state = {'__counter': '1', '__total_files': str(len(files))}

    x = findfunc.filenametotag(pattern, tag[PATH], True)
    emit('ftstatus', display_tag(x))

    bold_error = translate("Status Bar", "<b>{}</b>")

    try:
        newfilename = functions.move(tag, pattern, tag, state=state.copy())
        if newfilename:
            newfilename = newfilename['__path']
            emit('tfstatus', translate('Status Bar',
                                       "New Filename: <b>{}</b>"
                                       ).format(decode_fn(newfilename)))
        else:
            emit('tfstatus', '<b>No change</b>')
    except findfunc.ParseError as e:
        emit('tfstatus', bold_error.format(e.message))

    try:
        newfolder = functions.tag_dir(tag.tags, pattern, tag, state)
        if newfolder:
            newfolder = newfolder['__dirpath']
            dirstatus = translate('Dir Renaming',
                                  "Rename: <b>{}</b> to: <i>{}</i>"
                                  ).format(tag[DIRPATH], decode_fn(newfolder))
            emit('renamedirstatus', dirstatus)
    except findfunc.ParseError as e:
        emit('renamedirstatus', bold_error.format(e.message))

    selected = status['selectedtags']
    if not selected:
        emit('formatstatus', display_tag(''))
    else:
        selected = selected[0]
    try:
        val = tf(pattern, tag, state=state.copy())
        newtag = dict([(key, val) for key in selected])
        emit('formatstatus', display_tag(newtag))
    except findfunc.ParseError as e:
        emit('formatstatus', bold_error.format(e.message))


class _SignalObject(QObject):
    writeselected = pyqtSignal(object, name='writeselected')
    ftstatus = pyqtSignal(str, name='ftstatus')
    tfstatus = pyqtSignal(str, name='tfstatus')
    renamedirstatus = pyqtSignal(str, name='renamedirstatus')
    formatstatus = pyqtSignal(str, name='formatstatus')
    renamedirs = pyqtSignal(list, name='renamedirs')
    onetomany = pyqtSignal(dict, name='onetomany')
    renameselected = pyqtSignal(name='renameselected')
    adddock = pyqtSignal(str, object, int, name='adddock')
    highlight = pyqtSignal(list, name='highlight')
    writeaction = pyqtSignal([object, object, dict], name='writeaction')


obj = _SignalObject()
obj.emits = ['writeselected', 'ftstatus', 'tfstatus', 'renamedirstatus',
             'formatstatus', 'renamedirs', 'onetomany', 'renameselected',
             'adddock', 'highlight', 'writeaction']
obj.receives = [('filesselected', update_status),
                ('patternchanged', update_status)]


def emit_received(signal):
    def emit(*args):
        getattr(obj, signal).emit(*args)

    return emit


def emit(sig, *args):
    getattr(obj, sig).emit(*args)



import json
import re
import time
import urllib.error
import urllib.parse
import urllib.request
from collections import defaultdict
from itertools import chain
from xml.sax.saxutils import escape, quoteattr

from html.parser import HTMLParser
from xml.dom import minidom, Node

from puddlestuff.audioinfo import IMAGETYPES, get_mime, strlength
from puddlestuff.constants import CHECKBOX, COMBO
from puddlestuff.tagsources import (find_id, write_log, RetrievalError,
                          urlopen, parse_searchstring)
from puddlestuff.util import isempty, translate

SERVER = 'http://musicbrainz.org/ws/2/'

TEXT_NODE = Node.TEXT_NODE

ARTISTID = '#artistid'
ALBUMID = '#albumid'
LABELID = '#labelid'
INCLUDES = ''

ARTISTID_FIELD = 'mbrainz_artist_id'
ALBUMID_FIELD = 'mbrainz_album_id'

ARTIST_KEYS = {
    'name': 'artist',
    'sort-name': 'sortname',
    'id': ARTISTID_FIELD,
    'ext:score': '#score',
    'type': 'artist_type',
    'rating': 'mbrainz_rating',
}

ALBUM_KEYS = ARTIST_KEYS.copy()
ALBUM_KEYS.update({
    'name': 'album',
    'id': ALBUMID_FIELD,
    'type': 'album_type',
    'xml:ext': '#xml:ext',
    'title': 'album',
    'track-count': '__numtracks',
})
MEDIUM_KEYS = ALBUM_KEYS.copy()
MEDIUM_KEYS.update({
    'id': 'mbrainz_medium_id',
})

TRACK_KEYS = {
    'id': 'mbrainz_track_id',
    'position': 'track',
    'length': '__length',
    'rating': 'mbrainz_rating',
}

TO_REMOVE = ('recording', 'offset', 'count')

SMALL = 0
LARGE = 1
ORIG = 2

imagetypes = dict(reversed(i) for i in enumerate(IMAGETYPES))

mb_imagetypes = {
    "Front": "Cover (Front)",
    "Back": "Cover (Back)",
    "Other": "Other",
}


def children_to_text(node):
    if istext(node): return
    info = dict(list(node.attributes.items()))
    for ch in node.childNodes:
        if istext(ch): continue
        key = ch.tagName
        if key not in info:
            info[key] = node_to_text(ch)
        else:
            info[key] = to_list(info[key], node_to_text(ch))
    return info


def convert_dict(d, fm):
    return dict((fm.get(k, k), v) for k, v in d.items() if
                not isempty(v))


def fix_xml(xml):
    c = XMLEscaper()
    c.feed(album_xml)
    return c.xml


def istext(node):
    return getattr(node, 'nodeType', None) == TEXT_NODE


ESCAPE_CHARS_RE = re.compile(r'(?<!\\)(?P<char>[&|+\-!(){}[\]^"~*?:/])')


def solr_escape(value):
    r"""Escape un-escaped special characters and return escaped value.

    >>> solr_escape(r'foo+') == r'foo\+'
    True
    >>> solr_escape(r'foo\+') == r'foo\+'
    True
    >>> solr_escape(r'foo\\+') == r'foo\\+'
    True
    """
    #    value = value.replace('/', ' ')
    return ESCAPE_CHARS_RE.sub(r'\\\g<char>', value)


def node_to_text(node):
    if len(node.childNodes) > 1:
        return
    text_node = node.firstChild
    if istext(text_node):
        return text_node.data


def parse_album(xml):
    doc = minidom.parseString(xml)
    release_node = doc.getElementsByTagName('release')[0]

    return parse_release(release_node)


def parse_album_search(xml):
    doc = minidom.parseString(xml)
    nodes = doc.getElementsByTagName('release-list')[0].childNodes
    ret = []
    for i, node in enumerate(nodes):
        if istext(node):
            continue
        ret.append(parse_release(node))
    return ret


def parse_artist_credit(node):
    artists = parse_node(node, 'artist-credit', 'name-credit', 'artist')
    if not artists:
        return {}

    artist = ', '.join(z['artist']['name'] for z in artists)
    if len(artists) == 1:
        artist_id = artists[0]['artist']['id']
        return {
            'artist': artist,
            '#artist_id': artist_id,
            'mbrainz_artist_id': artist_id,
        }
    else:
        return {'artist': artist}


def parse_artist_relation(relations):
    ret = defaultdict(lambda: [])
    for r in to_list(relations['relation']):
        field = r['type']
        desc = ''

        if 'attribute-list' in r:
            desc = ', '.join(to_list(r['attribute-list']['attribute']))
        if 'artist' in r:
            if not desc:
                desc = r['artist']['name']
            else:
                if r['direction'] == 'backward':
                    field = '%s %s' % (desc, field)
                else:
                    field = '%s %s' % (field, desc)
                desc = r['artist']['name']
        if desc:
            ret[field].append(desc)
    return ret


def parse_artist_search(xml):
    doc = minidom.parseString(xml)
    nodes = doc.getElementsByTagName('artist-list')[0].childNodes
    ret = []
    for node in nodes:
        if istext(node):
            continue
        info = dict(list(node.attributes.items()))
        for ch in node.childNodes:
            if istext(node):
                continue
            info[ch.tagName] = node_to_text(ch)
        info = convert_dict(info, ARTIST_KEYS)
        info['#artist_id'] = info['mbrainz_artist_id']
        ret.append(info)
    return ret


def parse_label_list(release_node):
    labels = parse_node(release_node, 'label-info-list', 'label-info',
                        'label')

    catalogs = [z['catalog-number'] for z in labels if 'catalog-number' in z]
    label_names = [z['label']['name'] for z in labels
                   if 'label' in z and 'name' in z['label']]
    label_ids = [z['label']['id'] for z in labels
                 if 'label' in z and 'id' in z['label']]
    return {
        'label': label_names,
        'mbrainz_label_id': label_ids,
        'catalognumber': catalogs
    }


def parse_medium_list(r_node):
    mediums = parse_node(r_node, 'medium-list', 'medium', 'format')
    if not mediums:
        return {}

    mediums = [convert_dict(m, MEDIUM_KEYS) for m in mediums]
    info = mediums[0]
    info.update({'discs': str(len(mediums))})
    return info


def parse_node(node, header_tag, sub_tag, check_tag):
    ret = []
    nodes = [z for z in node.childNodes if
             getattr(z, "tagName", '') == header_tag]
    for node in nodes:
        info = children_to_text(node)
        for ch in node.getElementsByTagName(sub_tag):
            if ch not in node.childNodes:
                continue
            info = info.copy()
            info.update(rec_children_to_text(ch))
            if check_tag not in info:
                continue
            ret.append(info)
    return ret


def parse_recording_relation(relations):
    info = defaultdict(lambda: [])

    for relation in to_list(relations['relation']):
        recording = relation['recording']
        desc = None

        if 'artist-credit' in recording:
            artists = []
            for cr in to_list(recording['artist-credit']['name-credit']):
                if 'join-phrase' in cr:
                    artists.append(cr['join-phrase'])
                artists.append(cr['artist']['name'])

            unique_artists = []
            for z in artists:
                if z not in unique_artists:
                    unique_artists.append(z)

            desc = ' '.join(unique_artists)

        if 'title' in recording:
            if desc:
                desc = recording['title'] + ' by ' + desc
            else:
                desc = recording['title']
        if desc is not None:
            info[relation['type']].append(desc)
    return info


def parse_release(node):
    info = children_to_text(node)
    info.update(parse_artist_credit(node))

    info.update(parse_label_list(node))
    info.update(parse_medium_list(node))
    info = convert_dict(info, ALBUM_KEYS)
    info['#album_id'] = info['mbrainz_album_id']

    for k in ['count', 'track-list']:
        if k in info:
            del info[k]

    if 'disambiguation' in info:
        info['album'] = "%s (%s)" % (info['album'], info['disambiguation'])
        del (info['disambiguation'])

    tracks = []
    for medium in node.getElementsByTagName('medium'):
        tracks.extend(parse_track_list(medium))
    return info, tracks


def parse_track_list(node):
    tracks = []
    for i, t in enumerate(parse_node(node, 'track-list', 'track', 'position')):
        track = t['recording']
        rem_keys = set(track).union(TO_REMOVE)
        track.update((k, v) for k, v in t.items() if k not in rem_keys)

        if 'puid-list' in track:
            track['musicip_puid'] = track['puid-list']['id']
            del (track['puid-list'])

        if not isempty(track.get('relation-list')):
            for r in to_list(track['relation-list']):
                track.update(parse_track_relation(r))

        feat = to_list(track.get('artist-credit', {}).get('name-credit'))
        if feat:
            names = [(z['artist']['name'], z.get('joinphrase', ''))
                     for z in feat]

            track['artist'] = ''.join('%s%s' % a for a in names)

        for k, v in list(track.items()):
            if not isinstance(track[k], (str, list)):
                del (track[k])
            elif isinstance(v, list) and not isinstance(v[0], str):
                del (track[k])

        if 'length' in track:
            track['length'] = strlength(int(track['length']) / 1000)

        tracks.append(convert_dict(track, TRACK_KEYS))
    return tracks


def parse_track_relation(relation):
    if relation['target-type'] == 'recording':
        return parse_recording_relation(relation)
    elif relation['target-type'] == 'artist':
        return parse_artist_relation(relation)
    return {}


def rec_children_to_text(node):
    if istext(node): return
    info = dict(list(node.attributes.items()))
    for ch in node.childNodes:
        if istext(ch):
            continue
        text = node_to_text(ch)
        tag = ch.tagName
        if text is not None:
            info[tag] = to_list(info[tag], text) if tag in info else text
        elif ch.childNodes:
            v = rec_children_to_text(ch)
            info[tag] = to_list(info[tag], v) if tag in info else v
        elif ch.attributes:
            for k, v in ch.attributes.items():
                info[k] = to_list(info[k], v) if k in info else v
    return info


def retrieve_album(album_id):
    url = SERVER + 'release/' + album_id + \
          '?inc=recordings+artist-credits+puids+isrcs+tags+ratings' \
          '+artist-rels+recording-rels+release-rels+release-group-rels' \
          '+url-rels+work-rels+recording-level-rels+work-level-rels'

    data = urlopen(url)
    return parse_album(data)


def retrieve_cover_links(album_id, extra=None):
    if extra is None:
        url = "http://coverartarchive.org/release/" + album_id
    else:
        url = "http://coverartarchive.org/release/%s/%s" % (album_id, extra)
    write_log(translate('MusicBrainz', "Retrieving cover: {}").format(url))
    try:
        data, code = urlopen(url, code=True)
    except RetrievalError as e:
        if e.code == 404:
            raise RetrievalError(translate("MusicBrainz",
                                           "No images exist for this album."), 404)
        raise e

    if code == 200:
        if extra is None:
            return json.loads(data)
        else:
            return data
    elif code == 400:
        raise RetrievalError(translate("MusicBrainz", "Invalid UUID"))
    elif code in (405, 406):
        raise RetrievalError(translate("MusicBrainz", "Invalid query sent."))
    elif code == 503:
        raise RetrievalError(translate("MusicBrainz",
                                       "You have exceeded your rate limit."))
    elif code == 404:
        raise RetrievalError(translate("MusicBrainz",
                                       "Image does not exist."))


def retrieve_covers(cover_links, size=LARGE):
    ret = []
    for cover in cover_links['images']:
        desc = cover.get('comment', "")
        cover_type = cover['types'][0]
        if cover_type in mb_imagetypes:
            cover_type = imagetypes[mb_imagetypes[cover_type]]
        else:
            cover_type = imagetypes["Other"]
        if cover == SMALL:
            image_url = cover['thumbnails']['small']
        elif cover == LARGE:
            image_url = cover['thumbnails']['large']
        else:
            image_url = cover['image']

        write_log(translate('MusicBrainz', "Retrieving image {}").format(image_url))
        image_data = urlopen(image_url)

        ret.append({'desc': desc, 'mime': get_mime(image_data),
                    "imagetype": cover_type, "data": image_data})

    return ret


def retrieve_front_cover(album_id):
    data = retrieve_cover_links(album_id, "front")
    return {'data': data, 'mime': get_mime(data)}


def search_album(album=None, artist=None, limit=25, offset=0, own=False):
    if own:
        if isinstance(album, str):
            album = solr_escape(album)

        return SERVER + 'release/?query=' + urllib.parse.quote_plus(album) + \
               '&limit=%d&offset=%d' % (limit, offset)

    if artist:
        query = 'artistname:' + urllib.parse.quote_plus(solr_escape(artist))

    if album:
        if isinstance(album, str):
            album = solr_escape(album)
        if artist:
            query = 'release:' + urllib.parse.quote_plus(album) + \
                    '%20AND%20' + query
        else:
            query = 'release:' + urllib.parse.quote_plus(album)

    return SERVER + 'release/?query=' + query.replace('%3A', '') + \
           '&limit=%d&offset=%d' % (limit, offset)


def search_artist(artist, limit=25, offset=0):
    query = urllib.parse.urlencode({
        'query': solr_escape(artist),
        'limit': limit,
        'offset': offset,
    })
    return SERVER + 'artist?' + query.replace('%3A', '')


def to_list(v, arg=None):
    if isinstance(v, list):
        if arg is not None:
            v.append(arg)
        return v
    else:
        return [v, arg] if arg is not None else [v]


class XMLEscaper(HTMLParser):

    def reset(self):
        HTMLParser.reset(self)
        self._xml = []

    def handle_data(self, data):
        self._xml.append(escape(data))

    def unknown_starttag(self, tag, attributes):
        attrib_str = ' '.join('%s=%s' % (k, quoteattr(v))
                              for k, v in attributes)
        self._xml.append('<%s %s>' % (tag, attrib_str))

    def unknown_endtag(self, tag):
        self._xml.append('</%s>' % tag)

    @property
    def xml(self):
        return ''.join(self._xml)


class MusicBrainz(object):
    name = 'MusicBrainz'

    group_by = ['album', 'artist']

    def __init__(self):
        super(MusicBrainz, self).__init__()
        self.__lasttime = time.time()
        self.__image_size = LARGE
        self.__num_images = 0
        self.__get_images = True

        self.preferences = [
            [translate('MusicBrainz', 'Retrieve Cover'), CHECKBOX, True],
            [translate('MusicBrainz', 'Cover size to retrieve:'), COMBO,
             [[translate('Amazon', 'Small'),
               translate('Amazon', 'Large'),
               translate('Amazon', 'Original Size')], 1]],
            [translate('MusicBrainz', 'Amount of images to retrieve:'), COMBO,
             [[translate('MusicBrainz', 'Just the front cover'),
               translate('MusicBrainz', 'All (can take a while)')], 0]],
        ]

    def keyword_search(self, s):
        if s.startswith(':a'):
            artist_id = s[len(':a'):].strip()
            try:
                url = search_album('arid:' +
                                   solr_escape(artist_id), limit=100, own=True)
                return parse_album_search(urlopen(url))
            except RetrievalError as e:
                msg = translate('MusicBrainz',
                                "<b>Error:</b> While retrieving {}: {}"
                                ).format(artist_id, escape(e))
                write_log(msg)
                raise
        elif s.startswith(':b'):
            r_id = s[len(':b'):].strip()
            try:
                return [self.retrieve(r_id)]
            except RetrievalError as e:
                msg = translate('MusicBrainz',
                                "<b>Error:</b> While retrieving Album ID {} ({})"
                                ).format(r_id, escape(e))
                write_log(msg)
                raise
        else:
            try:
                params = parse_searchstring(s)
            except RetrievalError as e:
                return parse_album_search(urlopen(search_album(s, limit=100)))
            if not params:
                return
            artist = params[0][0]
            album = params[0][1]
            return self.search(album, [artist], 100)

    def search(self, album, artists='', limit=40):
        if time.time() - self.__lasttime < 1000:
            time.sleep(1)

        ret = []
        check_matches = False
        if isempty(artists):
            artist = None
        if len(artists) > 1:
            artist = 'Various Artists'
        elif artists:
            if hasattr(artists, 'items'):
                artist = list(artists.keys())[0]
            else:
                artist = artists[0]

        if not album and not artist:
            raise RetrievalError('Album or Artist required.')

        write_log('Searching for %s' % album)

        if hasattr(artists, "items"):
            album_id = find_id(chain(*list(artists.values())), "mbrainz_album_id")
            if album_id:
                try:
                    write_log(translate('MusicBrainz',
                                        "Found album id {} in tracks. Retrieving").format(album_id))
                    return [retrieve_album(album_id)]
                except RetrievalError as e:
                    msg = translate('MusicBrainz',
                                    "<b>Error:</b> While retrieving Album ID {} ({})"
                                    ).format(album_id, escape(e))
                    write_log(msg)

        try:
            xml = urlopen(search_album(album, artist, limit))
        except urllib.error.URLError as e:
            write_log('Error: While retrieving search page %s' %
                      str(e))
            raise RetrievalError(str(e))
        write_log('Retrieved search results.')
        self.__lasttime = time.time()
        return parse_album_search(xml)

    def retrieve(self, albuminfo):
        try:
            album_id = albuminfo['#album_id']
        except TypeError:
            album_id = albuminfo
        if time.time() - self.__lasttime < 1000:
            time.sleep(1)
        ret = retrieve_album(album_id)
        self.__lasttime = time.time()
        image = self.retrieve_covers(album_id)
        if image:
            ret[0]['__image'] = image
        return ret

    def retrieve_covers(self, album_id):
        if not self.__get_images:
            return []
        if self.__num_images == 0:
            try:
                image = retrieve_front_cover(album_id)
                if image:
                    return [image]
            except RetrievalError as e:
                import traceback
                traceback.print_exc()
                print()
                write_log(translate('MusicBrainz',
                                    "Error retrieving image: {}").format(str(e)))
                return []
        else:
            return retrieve_covers(album_id, self.__image_size)

    def applyPrefs(self, args):
        self.__get_images = args[0]
        self.__image_size = args[1]
        self.__num_images = args[2]


info = MusicBrainz

if __name__ == '__main__':
    # retrieve_album('f504ebe7-8fb4-40e5-aa55-b6384bdf863e')
    # c = MusicBrainz()
    xml = open('/home/keith/Desktop/mb.xml', 'r').read()
    # x = c.search('New Again', 'Taking Back Sunday')
    tracks = parse_album(xml)[1]
    for z in tracks:
        print(z['title'], z['track'])



import os
import pickle
import time
import traceback
from collections import defaultdict
from copy import deepcopy
from functools import partial

import quodlibet.config
from PyQt5.QtCore import QDir, Qt
from PyQt5.QtWidgets import QCompleter, QFileDialog, QFileSystemModel, QHBoxLayout, QLabel, QLineEdit, QPushButton, QVBoxLayout, QWidget
from quodlibet.query import Query

from .. import audioinfo
from ..audioinfo.tag_versions import tags_in_file
from ..audioinfo.util import (del_deco, fn_hash, getdeco,
                              isempty, keys_deco, lngfrequency, lnglength, lngtime, set_total, setdeco,
                              stringtags, strlength, unicode_list, CaselessDict, MockTag)
from ..constants import HOMEDIR
from ..musiclib import MusicLibError
from ..util import to_string, translate

model_tag = audioinfo.model_tag

ATTRIBUTES = ['length', 'accessed', 'size', 'created',
              'modified', 'filetype']


def strbitrate(bitrate):
    """Returns a string representation of bitrate in kb/s."""
    return str(bitrate / 1000) + ' kb/s'


def strtime(seconds):
    """Converts UNIX time(in seconds) to more Human Readable format."""
    return time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(seconds))


time_fields = ['~#added', '~#lastplayed', '~#laststarted']
timefunc = lambda key: lambda value: {'__%s' % key[2:]: strtime(value)}
mapping = dict([(key, timefunc(key)) for key in time_fields])

mapping.update({
    'tracknumber': lambda value: {'track': [value]},
    '~#bitrate': lambda value: {'__bitrate': strbitrate(value)},
    '~#length': lambda value: {'__length': strlength(value)},
    '~#playcount': lambda value: {'playcount': [str(value)]},
    '~#rating': lambda value: {'rating': [str(value)]},
    '~#skipcount': lambda value: {'__skipcount': [str(value)]},
    '~mountpoint': lambda value: {'__mountpoint': value},
    '~#mtime': lambda value: {'__modified': strtime(value)},
    '~picture': lambda value: {'__picture': value},
})

time_fields = ['__added', '__lastplayed', '__laststarted', ]
timefunc = lambda key: lambda value: {'~#%s' % key[2:]: lngtime(value)}
revmapping = dict([(key, timefunc(key)) for key in time_fields])
revmapping.update({
    'track': lambda value: {'tracknumber': value},
    '__bitrate': lambda value: {'~#bitrate': lngfrequency(value)},
    '__length': lambda value: {'~#length': lnglength(value)},
    'playcount': lambda value: {'~#playcount': int(value)},
    'rating': lambda value: {'~#rating': float(value)},
    '__skipcount': lambda value: {'~#skipcount': int(value)},
    '__mountpoint': lambda value: {'~mountpoint': value},
    '__modified': lambda value: {'~#mtime': lngtime(value)},
    '__picture': lambda value: {'~picture': to_string(value)},
})


class Tag(MockTag):
    """Use as base for all tag classes."""
    mapping = audioinfo.mapping.get('puddletag', {})
    revmapping = audioinfo.revmapping.get('puddletag', {})
    IMAGETAGS = ()

    def __init__(self, libclass, libtags):
        MockTag.__init__(self)

        self.__tags = CaselessDict()
        tags = self.__tags

        tags.update(self.load(libtags['~filename'])[0])
        tags['__tag_read'] = 'QuodLibet'

        self.library = libclass
        self.remove = partial(libclass.delete, track=libtags)
        self._libtags = libtags

        for key, value in libtags.items():
            if not value and not isinstance(value, int):
                continue
            if key in mapping:
                tags.update(mapping[key](value))
            else:
                if not isinstance(value, str):  # Strings
                    try:
                        value = str(value, 'utf8', 'replace')
                    except (TypeError, ValueError):
                        try:
                            value = str(value)  # Usually numbers
                        except:
                            traceback.print_exc()
                            continue
                tags[key] = [value]
        del (tags['~filename'])

        self.filepath = libtags['~filename']
        self.set_attrs(ATTRIBUTES, self.__tags)
        self.update_tag_list()

    @property
    def filepath(self):
        return MockTag.get_filepath(self)

    @filepath.setter
    def filepath(self, val):
        self.__tags.update(MockTag.set_filepath(self, val))

    images = property(lambda self: [], lambda self, value: None)

    def __contains__(self, key):
        if self.revmapping:
            key = self.revmapping.get(key, key)
        return key in self.__tags

    def __deepcopy__(self, memo=None):
        tag = Tag(self._libclass, self._libtags)
        tag.update(deepcopy(self.__tags))
        return tag

    @del_deco
    def __delitem__(self, key):
        if key.startswith('__'):
            return
        else:
            if key == 'track':
                del (self._libtags['tracknumber'])
                del (self.__tags['track'])
            else:
                if key in self._libtags:
                    del (self._libtags[key])
                    del (self.__tags[key])

    @getdeco
    def __getitem__(self, key):
        return self.__tags[key]

    @setdeco
    def __setitem__(self, key, value):
        if key.startswith('__'):
            if key == '__total':
                set_total(self, value)
            elif key in fn_hash:
                setattr(self, fn_hash[key], value)
        elif isempty(value):
            if key in self:
                del (self[key])
            else:
                return
        else:
            self.__tags[key] = unicode_list(value)

    def delete(self):
        raise NotImplementedError

    @keys_deco
    def keys(self):
        return list(self.__tags.keys())

    def save(self, justrename=False):
        libtags = self._libtags
        tags = self.__tags
        newartist = to_string(tags.get('artist', ['']))
        oldartist = libtags.get('artist', '')

        newalbum = to_string(tags.get('album', ['']))
        oldalbum = libtags.get('album', '')

        self._libtags.update(self._tolibformat())
        self._libtags.write()
        if (newartist != oldartist) or (newalbum != oldalbum):
            self.library.update(oldartist, oldalbum, libtags)
        self.library.edited = True
        self.update_tag_list()

    def _tolibformat(self):
        libtags = {}
        tags = list(stringtags(self.__tags).items())
        for key, value in tags:
            if key in revmapping:
                libtags.update(revmapping[key](value))
            elif key.startswith('__'):
                continue
            else:
                libtags[key] = value

        libtags['~filename'] = self.filepath
        return libtags

    def update_tag_list(self):
        l = tags_in_file(self.filepath)
        if l:
            self.__tags['__tag'] = 'QuodLibet, ' + ', '.join(l)
        else:
            self.__tags['__tag'] = 'QuodLibet'


Tag = audioinfo.model_tag(Tag)


class QuodLibet(object):
    def __init__(self, filepath):
        self.edited = False
        self._tracks = pickle.load(open(filepath, 'rb'))

        try:
            quodlibet.config.init()
        except ValueError:
            pass

        self._filepath = filepath
        cached = defaultdict(lambda: defaultdict(lambda: []))
        for track in self._tracks:
            cached[track.get('artist', '')][track.get('album', '')
            ].append(track)
        self._cached = cached

    def get_tracks(self, maintag, mainval, secondary=None, secvalue=None):

        exists = lambda t: os.path.exists(t['~filename'])

        if secondary and secvalue:
            if secondary == 'album' and maintag == 'artist':
                tracks = (Tag(self, track) for track in
                          self._cached[mainval][secvalue] if exists(track))
                return [_f for _f in tracks if _f]
            else:
                def getvalue(track):
                    if (track.get(maintag) == mainval) and (
                            track.get(secondary) == secvalue):
                        return Tag(self, track)
        else:
            if maintag == 'artist':
                tracks = []
                [tracks.extend(z) for z in self._cached[mainval].values()]
                return [Tag(self, track) for track in tracks if
                        exists(track)]

            def getvalue(track):
                if track.get(maintag) == mainval:
                    return Tag(self, track)

        return list(filter(getvalue, self._tracks))

    def distinct_values(self, field):
        return set([track.get(field, '') for track in self._tracks])

    def distinct_children(self, parent, value, child):
        return set([track.get(child, '') for track in
                    self._tracks if track.get(parent, '') == value])

    @property
    def artists(self):
        return list(self._cached.keys())

    def get_albums(self, artist):
        return list(self._cached[artist].keys())

    def save(self):
        if not self.edited:
            return
        filepath = self._filepath + '.puddletag'
        pickle.dump(self._tracks, open(filepath, 'wb'))
        os.rename(self._filepath, self._filepath + '.bak')
        os.rename(filepath, self._filepath)

    def delete(self, track):
        artist = to_string(track.get('artist', ''))
        album = to_string(track.get('album', ''))
        self._cached[artist][album].remove(track)
        self._tracks.remove(track)
        if not self._cached[artist][album]:
            del (self._cached[artist][album])
            if not self._cached[artist]:
                del (self._cached[artist])
        self.edited = True

    def tree(self):
        title = 'title'
        for artist in self.artists:
            print(artist)
            albums = self.get_albums(artist)
            for album in albums:
                print('  ', album)
                tracks = self.get_tracks('artist', artist, 'album', album)
                for track in tracks:
                    print('      ', track[title][0] if title in track else '')

    def close(self):
        pass

    def search(self, text):
        try:
            filt = Query(text).search
        except Query.error:
            return []
        else:
            return [Tag(self, track) for track in filter(filt, self._tracks)]

    def update(self, artist, album, track):
        cached = self._cached
        try:
            cached[artist][album].remove(track)
        except ValueError:
            pass
        if not cached[artist][album]:
            del (cached[artist][album])
        if not cached[artist]:
            del (cached[artist])
        trackartist = track.get('artist', '')
        trackalbum = track.get('album', '')
        cached[trackartist][trackalbum].append(track)


class DirLineEdit(QLineEdit):
    def __init__(self, *args, **kwargs):
        super(DirLineEdit, self).__init__(*args, **kwargs)
        completer = QCompleter()
        completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)

        model = QFileSystemModel()
        model.setRootPath('/')
        model.setFilter(QDir.Filter.AllEntries | QDir.Filter.NoDotAndDotDot | QDir.Filter.Hidden)
        completer.setModel(model)

        self.setCompleter(completer)


class InitWidget(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.dbpath = DirLineEdit(os.path.join(HOMEDIR, ".quodlibet/songs"))
        self.configpath = DirLineEdit(os.path.join(HOMEDIR, ".quodlibet/config"))

        vbox = QVBoxLayout()

        def label(text, control):
            l = QLabel(text)
            l.setBuddy(control)
            return l

        vbox.addWidget(label(translate("QuodLibet", '&Library Path'), self.dbpath))

        hbox = QHBoxLayout()
        select_db = QPushButton(translate("QuodLibet", "..."))
        select_db.clicked.connect(self.select_db)
        hbox.addWidget(self.dbpath)
        hbox.addWidget(select_db)
        vbox.addLayout(hbox)

        vbox.addStretch(1)
        self.setLayout(vbox)

    def select_db(self):
        selectedFile = QFileDialog.getOpenFileName(self,
                                                   translate("QuodLibet", 'Select QuodLibet library file...'),
                                                   self.dbpath.text())
        filename = selectedFile[0]
        if filename:
            self.dbpath.setText(filename)

    def library(self):
        dbpath = self.dbpath.text()
        try:
            return QuodLibet(dbpath)
        except (IOError, OSError) as e:
            raise MusicLibError(0, translate(
                'QuodLibet', "{} ({})").format(e.strerror, e.filename))
        except (pickle.UnpicklingError, EOFError):
            raise MusicLibError(0, translate('QuodLibet',
                                             "{} is an invalid QuodLibet music library file."
                                             ).format(dbpath))


name = 'Quodlibet'

if __name__ == '__main__':
    lib = QuodLibet('/home/keith/.quodlibet/songs')
    artists = lib.artists
    import random

    d = defaultdict(lambda: defaultdict(lambda: []))
    for nothing in range(20):
        artist = artists[random.randint(0, len(artists))]
        for album, tracks in lib._cached[artist].items():
            d[artist][album] = [Tag(lib, t).usertags for t in tracks]
    import pprint

    f = {}
    for z in d:
        f[z] = dict(d[z])
    pprint.pprint(f)



# Copyright (C) 2008-2009 concentricpuddle

# This file is part of puddletag, a semi-good music tag editor.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os

from puddlestuff.constants import TAGSOURCE, HOMEDIR, TEXT
from puddlestuff.puddleobjects import gettags, getfiles, PuddleConfig

# print 'example2'

properties = {'type': TAGSOURCE}

# musicdir = '/mnt/variable/Music'
# dirs = [unicode(z, 'utf8') for z in os.listdir(musicdir)]
try:
    image = [{'data': open(os.path.join(HOMEDIR, 'flux/image.jpg'), 'rb').read()}]
except:
    image = {}
    raise


def equal(audio1, audio2, play=False, tags=('artist', 'album')):
    for key in tags:
        if (key in audio1) and (key in audio2):
            if ''.join(audio1[key]).lower() != ''.join(audio2[key]).lower():
                return False
        else:
            return False
    if play and 'play' not in audio2:
        return False
    return True


class Example(object):
    name = 'Old'
    group_by = ['artist', 'album']
    counter = 0

    def __init__(self):
        object.__init__(self)
        cparser = PuddleConfig()
        musicdir = cparser.get('exampletagsource', 'musicdir', HOMEDIR)
        self.preferences = [['Music &Directory: ', TEXT, musicdir]]
        self.applyPrefs([musicdir])

    def search(self, artist, albums):
        ret = []
        matches = {}
        # set_status('Searching %s - %s' % (artist, albums[0]))
        # #write_log('Retrieving %s %s' % (artist, albums[0]))
        albumtuple = [z.split(' - ', 1) for z in self._dirs
                      if z.startswith(artist) and ' - ' in z]
        albumtuple = [(i, z) for i, z in albumtuple]
        releases = []
        for z in albumtuple:
            if len(z) > 1:
                releases.append(z[1])
        # if not releases:
        # set_status('No releases found')
        # write_log('No releases found for %s' % artist)
        # else:
        # write_log('Found albums <b>%s</b>' % ', '.join(releases))
        matched = []
        low_releases = [z.lower() for z in releases]
        for album in albums:
            info = {'album': album, 'artist': artist}
            info['#extrainfo'] = ('Home Folder', 'file://' + HOMEDIR)
            if album.lower() in low_releases:
                if album.lower() in low_releases:
                    return [(info, [])]
        get_info = lambda album: {'artist': artist, 'album': album,
                                  '#extrainfo': ('Home Folder', 'file:///' + HOMEDIR)}
        return [(get_info(album), []) for album in releases]

    def retrieve(self, info):
        artist = info['artist']
        album = info['album']
        self.counter += 1
        dirpath = '%s/%s - %s' % (self.musicdir, artist, album)
        files = []
        for f in gettags(getfiles(dirpath)):
            if f:
                # tag = stringtags(f.usertags)
                # tag['__image'] = image
                files.append(f.usertags)
        return info, None

    def applyPrefs(self, values):
        musicdir = values[0]
        self.musicdir = musicdir
        cparser = PuddleConfig()
        cparser.set('exampletagsource', 'musicdir', musicdir)
        self.preferences[0][2] = musicdir
        isdir = os.path.isdir
        join = os.path.join
        self._dirs = [z for z in os.listdir(musicdir) if isdir(join(musicdir, z))]


info = [Example, None]
name = 'OldExample'

if __name__ == '__main__':
    x = Example()
    d = x.search('Angie Stone', ['The Very Best Of'])
    print(x.retrieve(d[0][0]))
